/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#include "Dijkstra.h"
#include "debug.h"

smerr_t Dijkstra::init()
{
  smerr_t rc;

  if(!_graph) return smeDIJ_NO_SUBWAYGRAPH;
  if(!_startStation) return smeDIJ_NO_START_STATION;
  if(!_destStation) return smeDIJ_NO_DEST_STATION;

  if((_startNodeIndex=_graph->add_startNodeForStation(_startStation))<0) {
	cerror << "start node not added rc=" << _startNodeIndex << endl;
	return _startNodeIndex;
  }

  if((_destNodeIndex=_graph->add_destNodeForStation(_destStation))<0) {
	cerror << "dest node not added rc=" << _destNodeIndex << endl;
	return _destNodeIndex;
  }

  
  // ---- initialize nodes and node set ----

  if((rc=init_notVisited())!=smeOK) {
	return rc;
  }

  for(int index=0; index<_graph->get_numNodes(); index++) {
	if(index!=_startNodeIndex) {
	  SGNode *node=_graph->get_nodeAt(index);
	  node->set_prevNode(NULL);
	  node->set_distance(INFINITE);
	  add_notVisited(index);
	}
  }

  SGNode *startNode=_graph->get_nodeAt(_startNodeIndex);
  if(!startNode) {
	cerror << "no start node" << endl;
	return smeDIJ_NO_START_NODE;
  }

  SGEdge *edge=startNode->first_edge();
  while(edge) {
	SGNode *target=edge->get_target();
	target->set_prevNode(startNode);
	target->set_distance(edge->get_weight());
	edge=startNode->next_edge();
  }

  return smeOK;
}

smerr_t Dijkstra::execute()
{
  bool destFound=false;

  SGNode *destNode=_graph->get_nodeAt(_destNodeIndex);

  while(!destFound) {
	SGNode *node=extract_nearestNode();
	if(!node) {
	  // nomore nodes in set
	  cerror << "destination " << _destNodeIndex << " not found" << endl;
	  return smeDIJ_DEST_NOT_FOUND;
	}

	SGEdge *edge=node->first_edge();
	while(edge) {
	  SGNode *target=edge->get_target();
	  if((node->get_distance()+(int)edge->get_weight())<target->get_distance()) {
		target->set_prevNode(node);
		target->set_distance(node->get_distance()+edge->get_weight());
	  }
	  if(target==destNode) {
		destFound=true;
	  }
	  edge=node->next_edge();
	}
  }
  return smeOK;
}

smerr_t Dijkstra::cleanup()
{
  cleanup_notVisited();

  if(_graph) {
	_graph->remove_tempNodes();
  }

  return smeOK;
}

smerr_t Dijkstra::get_pathNodes(ObjectList& list)
{
  
  SGNode *node=_graph->get_nodeAt(_destNodeIndex);
  while(node) {
	list.addToBegin(node);
	node=node->get_prevNode();
  }
  return smeOK;
}

smerr_t Dijkstra::init_notVisited()
{
  if(_setNotVisited) {
	delete [] _setNotVisited;
	_setNotVisited=NULL;
  }

  _setLength=_graph->get_numNodes()-1;
  if(_setLength<1) {
	return smeDIJ_ILLEGAL_SETLENGTH;
  }

  _setNotVisited=new int[_setLength];
  if(!_setNotVisited) {
	return smeOUT_OF_MEMORY;
  }

  reset_notVisited();

  return smeOK;
}

smerr_t Dijkstra::add_notVisited(int nodeindex)
{
  for(int i=0; i<_setLength; i++) {
	if(_setNotVisited[i]<0) {
	  _setNotVisited[i]=nodeindex;
	  return smeOK;
	}
  }

  return smeDIJ_SET_FULL;
}

smerr_t Dijkstra::reset_notVisited()
{
  for(int i=0; i<_setLength; i++) {
	_setNotVisited[i]=-1;
  }

  return smeOK;
}

smerr_t Dijkstra::cleanup_notVisited()
{
  if(_setNotVisited) {
	delete [] _setNotVisited;
	_setNotVisited=NULL;
  }

  return smeOK;
}

SGNode *Dijkstra::extract_nearestNode()
{
  int nearest=-1;
  int distance=INFINITE;
  int nodeindex=0;
  SGNode *node=NULL;

  // find node with shortest distance to start node
  for(int setindex=0; setindex<_setLength; setindex++) {
	nodeindex=_setNotVisited[setindex];
	if(nodeindex<0)
	  continue;

	node=_graph->get_nodeAt(nodeindex);
	if(!node) {
	  // better abort?
	  continue;
	}

	if(node->get_distance()<distance) {
	  nearest=setindex;
	  distance=node->get_distance();
	}
  }
  
  if(nearest==-1) {
	// no more nodes available?
	//printf("WAAH\n");
  	return NULL;
  }
  	// extrace nearest node from set of not visited nodes
  nodeindex=_setNotVisited[nearest];
  _setNotVisited[nearest]=-2;
  node=_graph->get_nodeAt(nodeindex);
 
  return node;
  
}





