/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#ifndef __EXCEPTIONS_H__
#define __EXCEPTIONS_H__

#include "baseclasses.h"
#include "SMError.h"


/*	=============================================
	Exception classes
	============================================= */

class ExceptionBase : public Object
{
 public:
  virtual ~ExceptionBase() {}

  String& get_errstr() { return _errstr; }
  int get_errnum() { return _errnum; }
  int get_extype() { return _extype; }

 protected:
  ExceptionBase() {
	_errstr="unknown";
	_errnum=0;
	_extype=smeGOT_EXCEPTION;;
  }

  ExceptionBase(const String& errstr, const int errnum) {
	_errstr=errstr;
	_errnum=errnum;
	_extype=smeGOT_EXCEPTION;
  }

  ExceptionBase(const ExceptionBase& e) : Object() {
	_errstr=e._errstr;
	_errnum=e._errnum;
	_extype=e._extype;
  }

  String _errstr;
  int  _errnum;
  smerr_t _extype;
};


class FileException : public ExceptionBase
{
 public:
  FileException() : ExceptionBase() {
	_filename="unknown";
	_extype=smeGOT_FILE_EXCEPTION;
  }
  FileException(const String& string, const int num, const String& filename) : ExceptionBase(string, num) {
	_filename=filename;
	_extype=smeGOT_FILE_EXCEPTION;
  }
  FileException(const FileException& e) : ExceptionBase(e) {
	_filename=e._filename;
  };
  virtual ~FileException() {
	_filename=NULL;
  };
  
  String& get_filename() { return _filename; }

 protected:
  String _filename;
};

class MemoryException : public ExceptionBase
{
 public:
  MemoryException() : ExceptionBase() {
	_extype=smeGOT_MEMORY_EXCEPTION;
  }
  MemoryException(const String& string, const int num) : ExceptionBase(string, num) {
	_extype=smeGOT_MEMORY_EXCEPTION;
  };
  MemoryException(const FileException& e) : ExceptionBase(e) {};
  virtual ~MemoryException() {};
 protected:
};


/*	=============================================
	Exception classes
	============================================= */
class XMLException : public ExceptionBase
{
 public:
  XMLException(const String& errstr, const int num) : ExceptionBase(errstr, num),
	_xmlTag(NULL), _line(0), _column(0) {
	_extype=smeGOT_XML_EXCEPTION;
  }

  XMLException(const String& errstr, const int num, const int line, const int column) : 
	ExceptionBase(errstr, num), 
	_xmlTag("none"), _line(line), _column(column) {
	_extype=smeGOT_XML_EXCEPTION;
  }

  XMLException(const String& errstr, const int num, const String& xmlTag, 
			   const int line, const int column) :
	ExceptionBase(errstr, num),
	_xmlTag(xmlTag), _line(line), _column(column) {
	_extype=smeGOT_XML_EXCEPTION;
  }
	

  XMLException(const XMLException& e) : ExceptionBase(e) {
	_xmlTag=e._xmlTag;
	_line=e._line;
	_column=e._column;
  }
  // do not call this...

  XMLException() : ExceptionBase(),
	_xmlTag(NULL), _line(0), _column(0) {
	_extype=smeGOT_XML_EXCEPTION;
  }


  virtual ~XMLException() {};

  int get_line() { return _line; }
  int get_column() { return _column; }
  String& get_xmlTag() { return _xmlTag; }

 protected:
  String _xmlTag;
  int _line;
  int _column;
  
};


class SubwayMapException : public ExceptionBase
{
 public:
  SubwayMapException() : ExceptionBase() {}
  SubwayMapException(const String& string, const int num) : ExceptionBase(string, num) {};
  SubwayMapException(const SubwayMapException& e) : ExceptionBase(e) {};
  virtual ~SubwayMapException() {};
 protected:
};

class SubwayMapInvalidStation : public ExceptionBase
{
 public:
  SubwayMapInvalidStation() : ExceptionBase() {
	_extype=smeGOT_INVALID_STATION_EXCEPTION;
  }
  SubwayMapInvalidStation(String& id) : ExceptionBase(), _id(id) {
	_errstr="invalid station";
	_errnum=999;
	_extype=smeGOT_INVALID_STATION_EXCEPTION;
  }
  virtual ~SubwayMapInvalidStation() {
	_id=NULL;
  }
  String& get_id() { return _id; }

 protected:
  String _id;
};
class SubwayMapLineNotFound : public ExceptionBase
{
 public:
  SubwayMapLineNotFound() : ExceptionBase(), _id(NULL) {}
  SubwayMapLineNotFound(String& id) : ExceptionBase(), _id(id) {
	_errstr="invalid line";
	_errnum=998;
	_extype=smeGOT_LINE_NOT_FOUND_EXCEPTION;
  }
  virtual ~SubwayMapLineNotFound() {
	_id=NULL;
  }
  String& get_id() { return _id; }

 protected:
  String _id;
};
class SubwayMapNotOnSameLine : public ExceptionBase
{
 public:
  SubwayMapNotOnSameLine() : ExceptionBase() {}
  SubwayMapNotOnSameLine(String& line, String& station1, String& station2) :
	ExceptionBase() {
	_errnum=smeROUTE_NOT_ON_SAME_ROUTE;
	_line=line;
	_station1=station1;
	_station2=station2;
	_extype=smeGOT_NOT_ON_SAME_LINE_EXCEPTION;
  }
  String& get_line() { return _line; }
  String& get_station1() { return _station1; }
  String& get_station2() { return _station2; }
 protected:
  String _line;
  String _station1;
  String _station2;
};



#endif
