
/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#include "RouteFinder.h"
#include "Dijkstra.h"
#include "debug.h"


/*	=================================================
	Parameter initialization
	================================================= */


void RouteFinder::use_subwayMap(SubwayMap *map)
{
  if(_subwayGraph) {
	delete _subwayGraph;
	_subwayGraph=NULL;
  }

  _subwayMap=map;
};

smerr_t RouteFinder::set_startStationByName(String& name)
{
  if(!_subwayMap) return smeROUTE_NO_SUBWAYMAP;

  ObjectListIterator it;

  _startStation=_subwayMap->find_stationByName(name, it);
  if(_startStation==NULL)
	return smeROUTE_STATION_NOT_FOUND;

  return smeOK;
}

smerr_t RouteFinder::set_destStationByName(String& name)
{
  if(!_subwayMap) return smeROUTE_NO_SUBWAYMAP;

  ObjectListIterator it;
  _destStation=_subwayMap->find_stationByName(name, it);
  if(_destStation==NULL)
	return smeROUTE_STATION_NOT_FOUND;

  return smeOK;
}

smerr_t RouteFinder::set_operationTime(OperationTime::EDay day, unsigned int hour) 
{
  _opDay=day;
  _opHour=hour;
  if(_subwayGraph) {
	// delete current subway graph. will be rebuilt
	// with new operation time parameters
	delete _subwayGraph;
	_subwayGraph=NULL;
  }

  return smeOK;
}



/*	=================================================
	START Route Finder
	================================================= */
smerr_t RouteFinder::findRoute()
{
  smerr_t rc;

  // Check start and destination
  if(!_startStation) return smeROUTE_NO_START_STATION;
  if(!_destStation) return smeROUTE_NO_DEST_STATION;
  if(!_subwayMap) return smeROUTE_NO_SUBWAYMAP;


  if(!_subwayGraph) {
	// no graph built yet...
	_subwayGraph=new SubwayGraph(_subwayMap);
	_subwayGraph->set_operationTime(_opDay, _opHour);
	_subwayGraph->build();
  } else {
	_subwayGraph->reset();
  }
 
  //cout << _subwayGraph << endl;
  // route finding starts here...

  Dijkstra *dijkstra=new Dijkstra(_subwayGraph);
  dijkstra->set_startStation(_startStation);
  dijkstra->set_destStation(_destStation);


  ObjectList nodeList;

  // ---- find shortest route ----
  rc=dijkstra->init();
  if(rc!=smeOK) {
	cerror << "dijkstra->init() failed, rc=" << rc << endl;
	delete dijkstra;
	return rc;
  }

  rc=dijkstra->execute();
  if(rc!=smeOK) {
	cerror << "dijkstra->execute() failed, rc=" << rc << endl;
	delete dijkstra;
	return rc;
  }
  cinfo << "route found" << endl;
  dijkstra->get_pathNodes(nodeList);

  Route *route=NULL;
  route=create_routeFromNodeList(nodeList);
  add_routeFound(route);

  bool haveDirectRoute=false;
  bool hasByFeetConnection=false;

  if(route->get_numConnections()<2) {
	haveDirectRoute=true;
  }

  hasByFeetConnection=route->has_byFeetConnection();

  route=NULL;
  nodeList.cleanup();

  if(!haveDirectRoute) {

	// --- try to find direct route ---
	_subwayGraph->reset(SubwayGraph::direct);

	rc=dijkstra->init();
	if(rc!=smeOK) {
	  cerror << "dijkstra->init() failed, rc=" << rc << endl;
	  delete dijkstra;
	  return rc;
	}
  
	rc=dijkstra->execute();
	if(rc!=smeOK) {
	  cerror << "dijkstra->execute() failed, rc=" << rc << endl;
	  delete dijkstra;
	  return rc;
	}
	cinfo << "route found" << endl;
	dijkstra->get_pathNodes(nodeList);
	route=create_routeFromNodeList(nodeList);

	if(route) {
	  if(route->get_numConnections()<2) {
		cinfo << "direct route found" << endl;
		add_routeFound(route);
	  } else {
		delete route;
	  }
	}

	route=NULL;
	nodeList.cleanup();
  }

  if(hasByFeetConnection) {
	// --- try to find a short route without the need to walk ---

	_subwayGraph->reset(SubwayGraph::dontwalk);

	rc=dijkstra->init();
	if(rc!=smeOK) {
	  cerror << "dijkstra->init() failed, rc=" << rc << endl;
	  delete dijkstra;
	  return rc;
	}
  
	rc=dijkstra->execute();
	if(rc!=smeOK) {
	  cerror << "dijkstra->execute() failed, rc=" << rc << endl;
	  delete dijkstra;
	  return rc;
	}
	cinfo << "route found" << endl;
	dijkstra->get_pathNodes(nodeList);
	route=create_routeFromNodeList(nodeList);

	if(route) {
	  if(route->has_byFeetConnection()==false) {
		cinfo << "route without walk found" << endl;
		add_routeFound(route);
	  } else {
		delete route;
	  }
	}

	route=NULL;
	nodeList.cleanup();

  }


  delete dijkstra;

  return 0;
}


Route *RouteFinder::create_routeFromNodeList(ObjectList& nodeList)
{
  ObjectListIterator it;
  Route *route=NULL;
  Connection *conn=NULL;
  int timeToChange=0;
  bool byFeet=false;

  if((route=new Route)==NULL) {
	cerror << "cannot create route object, out of memory" << endl;
	return NULL;
  }

  SGNode *node=(SGNode *)nodeList.first(it);
  while(node) {

	// cout << node << endl;

	if(node->get_type()==SGNode::startNode) {
	  node=(SGNode *)nodeList.next(it);
	  continue;
	}


	if(node->get_type()==SGNode::destNode) {
	  node=(SGNode *)nodeList.next(it); // should be NULL
	  continue;
	}

	// get the next node
	SGNode *nextNode=(SGNode *)nodeList.next(it);
	SGEdge *edge=node->find_edgeToTarget(nextNode);

	switch(edge->get_type()) {
	case SGEdge::changeToLine:
	  if(conn==NULL) {
		cerror << "changeToLine and conn==NULL: check map" << endl;
		break;
	  }
	  conn->set_toStation(node->get_station());
	  route->add_connection(conn);
	  conn=NULL;

	  timeToChange=edge->get_realweight();
	  byFeet=false;
	  break;

	case SGEdge::changeToStation:
	  if(conn) {
		// last node on this route before changing to another
		// station
		conn->set_toStation(node->get_station());
		route->add_connection(conn);
		conn=NULL;
	  }

	  conn=new Connection;
	  conn->set_fromStation(node->get_station());
	  conn->set_toStation(nextNode->get_station());
	  conn->set_byFeet(true);
	  conn->set_timeToChange(edge->get_realweight());
	  route->add_connection(conn);
	  conn=NULL;
	  break;

	case SGEdge::toDestNode:
	  // next node will be the temporary destination node
	  if(conn) {
		conn->set_toStation(node->get_station());
		route->add_connection(conn);
		conn=NULL;
	  }
	  break;

	default:	// normal edge
	  // no connection object, yet.
	  if(!conn) {
		// start station of connection
		conn=new Connection;
		conn->set_fromStation(node->get_station());
		conn->set_toStation(NULL);
		conn->set_useLine(node->get_line());
		conn->set_routeOfUsedLine(node->get_routeOfLine());
		conn->set_timeToChange(timeToChange);
		conn->set_byFeet(false);
		timeToChange=0;
	  } else {
		// passed station
		conn->add_passedStation(node->get_station());
	  }
	  
	  break;
	}

	// go on with next node
	node=nextNode;
  }

  return route;
}

void RouteFinder::add_routeFound(Route *route)
{
  _routes.add(route);
}

Route *RouteFinder::first_routeFound(ObjectListIterator& it)
{
  return (Route *)_routes.first(it);
}

Route *RouteFinder::next_routeFound(ObjectListIterator& it)
{
  return (Route *)_routes.next(it);
}

Route *RouteFinder::first_directRouteFound(ObjectListIterator& it)
{
  Route *tmpRoute=first_routeFound(it);

  while(tmpRoute && tmpRoute->get_numConnections()>1) {
	tmpRoute=next_routeFound(it);
  }
  return tmpRoute;
}

Route *RouteFinder::next_directRouteFound(ObjectListIterator& it)
{
  Route *tmpRoute=next_routeFound(it);

  while(tmpRoute && tmpRoute->get_numConnections()>1) {
	tmpRoute=next_routeFound(it);
  }
  return NULL;
}

Route *RouteFinder::get_shortestRouteFound()
{
  ObjectListIterator it;
  Route *shortest=NULL;
  Route *tmpRoute=NULL;
  SortRouteJourneyTimeUp sorter;
  
  _routes.bsort(sorter);

  
  tmpRoute=first_routeFound(it);
  while(tmpRoute) {
	if(shortest==NULL) shortest=tmpRoute;
	if(tmpRoute->calc_journeyTime()>shortest->calc_journeyTime()) return shortest;
	if(tmpRoute->get_numConnections() < shortest->get_numConnections())
	  shortest=tmpRoute;

	tmpRoute=next_routeFound(it);
  }

  return shortest;
}

Route *RouteFinder::get_shortestDirectRouteFound()
{
  ObjectListIterator it;
  Route *shortest=NULL;
  Route *tmpRoute=NULL;
  SortRouteConnectionsUp sorter;

  _routes.bsort(sorter);

  tmpRoute=first_routeFound(it);
  while(tmpRoute) {
	if(shortest==NULL) shortest=tmpRoute;
	if(tmpRoute->get_numConnections()>shortest->get_numConnections()) return shortest;
	int jt=tmpRoute->calc_journeyTime();
	if(jt<0) {
	  // Error with current route... what now?
	  tmpRoute=next_routeFound(it);
	  continue;
	}
	if(tmpRoute->calc_journeyTime()<shortest->calc_journeyTime())
	  shortest=tmpRoute;

	tmpRoute=next_routeFound(it);
  }
  

  return shortest;
}

Route *RouteFinder::get_directRouteFound()
{
  return NULL;
}

/*	========================================================== */






smerr_t RouteFinder::findAllDirectConnections()
{
  // Check start and destination
  if(!_startStation) return smeROUTE_NO_START_STATION;
  if(!_destStation) return smeROUTE_NO_DEST_STATION;
  if(!_subwayMap) return smeROUTE_NO_SUBWAYMAP;

  ObjectListIterator lineIt;
  Line *line=_subwayMap->first_line(lineIt);
  while(line) {
  	ObjectListIterator routeIt;
  	RouteOfLine *routeOfLine=line->first_routeOfLine(routeIt);
  	while(routeOfLine) {
		findConnectionOnRoute(routeOfLine);
		routeOfLine=line->next_routeOfLine(routeIt);
  	}
  	line=_subwayMap->next_line(lineIt);
  }
  return smeOK;  	
}


smerr_t RouteFinder::findConnectionOnRoute(RouteOfLine* routeOfLine)
{
  if(!routeOfLine)
  	return smeROUTE_MISSING_PARAMETER;
  	
  if(!routeOfLine->operates_at(_opDay, _opHour)) {
  	// doesn't operate now...
  	return smeOK;
  }
  
  if(!routeOfLine->hasStation(_startStation) || !routeOfLine->hasStation(_destStation)) {
  	return smeROUTE_NOT_ON_SAME_ROUTE;
  }
  
  // try forward
  ObjectListIterator stopIt;
  StopAt *stop=routeOfLine->find_stopByStation(_startStation, stopIt);
  if(!stop) {
  	return smeROUTE_ERROR;
  }
  bool found=false;
  while(stop && !found) {
  	if(stop->get_station()==_destStation) { // Ok, found
  		found=true;
  	}
  	stop=routeOfLine->next_stop(stopIt);
  }
  
  if(!found && routeOfLine->get_oneway()) {
  	// destination not found in forward direction and this route is oneway
  	return smeROUTE_NOT_ON_SAME_ROUTE;
  }
  
  if(found) {
	Route *route=new Route;
	Connection *conn=new Connection;
	conn->set_fromStation(_startStation);
	conn->set_toStation(_destStation);
	conn->set_useLine(routeOfLine->get_line());
	conn->set_routeOfUsedLine(routeOfLine);
	conn->set_timeToChange(0);
	conn->set_byFeet(false);
	route->add_connection(conn);
	add_routeFound(route); 	
	return smeOK;
  }
  
  if(!found) {
  	// try backwards
  	stop=routeOfLine->find_stopByStation(_startStation, stopIt);
  	while(stop && !found) {
  		if(stop->get_station()==_destStation) {
  			found=true;
  		}
  		stop=routeOfLine->prev_stop(stopIt);
  	}
  }
  if(found) {
	// must have found a connection...
	Route *route=new Route;
	Connection *conn=new Connection;
	conn->set_fromStation(_startStation);
	conn->set_toStation(_destStation);
	conn->set_useLine(routeOfLine->get_line());
	conn->set_routeOfUsedLine(routeOfLine);
	conn->set_timeToChange(0);
	conn->set_byFeet(false);
	route->add_connection(conn);
	add_routeFound(route);
	return smeOK;
  }
  
  return smeROUTE_ERROR;
}





smerr_t RouteFinder::set_startStation(Station* station)
{
  if(!station)	return smeROUTE_ERROR;
  if(!_subwayMap) return smeROUTE_NO_SUBWAYMAP;
		
  _startStation=station;
  return smeOK;
}


smerr_t RouteFinder::set_destStation(Station* station)
{
  if(!station) return smeROUTE_ERROR;
  if(!_subwayMap) return smeROUTE_NO_SUBWAYMAP;
	
  _destStation=station;
  return smeOK;
}