/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#ifndef __ROUTEFINDER_H__
#define __ROUTEFINDER_H__

#include "baseclasses.h"
#include "Exceptions.h"
#include "SubwayMap.h"
#include "SubwayGraph.h"
#include "RouteFinderBase.h"
#include "SMError.h"

/*	=============================================
	Classes for finding routes in subway maps
	============================================= */

class RouteFinder : public Object
{
 public:
  RouteFinder() : 	_startStation(NULL), _destStation(NULL), 
	_subwayMap(NULL), _subwayGraph(NULL), _exception(NULL) {
	
	// Set current time as operation time
	OperationTime::convert_time_t(_opDay, _opHour);

  }

  ~RouteFinder() {
	cleanup();
  }

  smerr_t set_startStationByName(String& name);
  smerr_t set_destStationByName(String& name);

  smerr_t set_startStation(Station* station);
  smerr_t set_destStation(Station* station);

  smerr_t set_operationTime(OperationTime::EDay day, unsigned int hour);

  void use_subwayMap(SubwayMap *map);
  smerr_t findRoute();

  smerr_t findAllDirectConnections();
  

  // Request of found routes
  Route *get_shortestRouteFound();
  Route *get_shortestDirectRouteFound();
  Route *get_directRouteFound();

  Route *first_routeFound(ObjectListIterator& it);
  Route *next_routeFound(ObjectListIterator& it);

  Route *first_directRouteFound(ObjectListIterator& it);
  Route *next_directRouteFound(ObjectListIterator& it);

  void cleanup_routesFound() { _routes.cleanup(1); }

  ObjectList& get_routes() { return _routes; }

  void cleanup() {
	_startStation=NULL;
	_destStation=NULL;
	_routes.cleanup(1);
	if(_subwayGraph) {
	  delete _subwayGraph;
	  _subwayGraph=NULL;
	}
  }

 protected:
	smerr_t findConnectionOnRoute(RouteOfLine* routeOfLine);
	Station *_startStation;
	Station *_destStation;

  void add_routeFound(Route *route);
  Route *create_routeFromNodeList(ObjectList& nodeList);

 protected:
  smerr_t init() {
	_startStation=NULL;
	_destStation=NULL;
	return smeOK;
  };

 protected:
  SubwayMap *_subwayMap;
  SubwayGraph *_subwayGraph;
  ObjectList _routes;

  OperationTime::EDay _opDay;
  unsigned int _opHour;

 protected:
  // -----------------------------------
  // Error handling/Exception simulation
  // -----------------------------------
  smerr_t exception(const ExceptionBase *e) { 
	_exception=const_cast<ExceptionBase *>(e); 
	return _exception->get_extype();
  }

  ExceptionBase *_exception;

 public:
  ExceptionBase *get_exception() { return _exception; }
  void clean_exception() {
	if(_exception) delete _exception;
	_exception=NULL;
  }

};




#endif
