/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/


#include "RouteFinderBase.h"

/*	==========================================================
	Route and Connection classes
	========================================================== */
Connection::Connection(Connection *conn)
{
  _fromStation=conn->_fromStation;
  _toStation=conn->_toStation;
  _useLine=conn->_useLine;
  _routeOfUsedLine=conn->_routeOfUsedLine;
  _connectionTime=conn->_connectionTime;
  _timeToChange=conn->_timeToChange;
  _byFeet=conn->_byFeet;
}


int Connection::calc_time(int recalc)
{
  if(recalc) _connectionTime=-1;

  if(_connectionTime!=-1)	// already calculated
	return _connectionTime;

  if(!_byFeet) {
	if(!_fromStation || !_toStation || !_useLine || !_routeOfUsedLine) {
	  return	smeROUTE_MISSING_PARAMETER;
	}


	_direction=_routeOfUsedLine->get_routeDirection(_fromStation, _toStation);

	if(_direction!=forward && _direction!=backwards) {
	  return smeROUTE_NOT_ON_SAME_ROUTE;
	}

	// _connectionTime=_timeToChange;
	_connectionTime=0;

	ObjectListIterator it;
	StopAt *stop=_routeOfUsedLine->find_stopByStation(_fromStation, it);

	while(stop && stop->get_station()!=_toStation) {
	  if(_direction==forward) {
		_connectionTime+=stop->get_timeToNext();
		stop=_routeOfUsedLine->next_stop(it);
	  }

	  if(_direction==backwards) {
		_connectionTime+=stop->get_timeToPrev();
		stop=_routeOfUsedLine->prev_stop(it);
	  }
	}
  } else {
	if(!_fromStation || !_toStation) {
	  return smeROUTE_MISSING_PARAMETER;
	}

	_connectionTime=_timeToChange;
  }
  
  return _connectionTime;
}

String& Connection::get_directionName()
{
  if(_direction==forward) {
	return _routeOfUsedLine->get_endstation();
  }
  if(_direction==backwards) {
	return _routeOfUsedLine->get_startstation();
  }
  if(_byFeet) {
	return _toStation->get_name();
  }

  return NULLString;
}

bool Connection::add_passedStation(Station *station)
{
  _passedStations.add(station);
  return true;
}

Station *Connection::first_passedStation()
{
  return (Station *)_passedStations.first(_passedStationsIt);
}

Station *Connection::next_passedStation()
{
  return (Station *)_passedStations.next(_passedStationsIt);
}


// Copy constructor
Route::Route(Route *route) : _journeyTime(-1)
{
  copyFrom(route);
}

void Route::add_connection(Connection *conn)
{
  _connections.add(conn);
}

Connection *Route::first_connection(ObjectListIterator& it)
{
  return (Connection *)_connections.first(it);
}

Connection *Route::next_connection(ObjectListIterator& it)
{
  return (Connection *)_connections.next(it);
}

int Route::get_numConnections() 
{
  return _connections.length();
}

bool Route::has_byFeetConnection()
{
  ObjectListIterator it;
  Connection *conn=(Connection *)_connections.first(it);
  while(conn) {
	if(conn->get_byFeet())
	  return true;
	conn=(Connection *)_connections.next(it);
  }
  return false;
}

int Route::calc_journeyTime(int withChanges)
{
  _journeyTime=0;
  ObjectListIterator it;
  Connection *conn=first_connection(it);
  while(conn) {
	int ct=conn->calc_time();
	if(ct<0) {
	  // set journey time to maximum value
	  _journeyTime=0xffff;
	  return ct;
	}
	_journeyTime+=ct;
	if(withChanges  && !conn->get_byFeet()) {
		_journeyTime+=conn->get_timeToChange();
	}
	conn=next_connection(it);
  }
  return _journeyTime;
}

void Route::copyFrom(Route *route)
{
  cleanup();

  if(route==NULL) {
	return;
  }

  _journeyTime=route->_journeyTime;

  ObjectListIterator it;
  Connection *conn=route->first_connection(it);
  while(conn) {
	Connection *conn2=new Connection(conn);
	add_connection(conn2);

	conn=route->next_connection(it);
  }
}


/*	==============================================
	Route list sorting classes
	============================================== */

int SortRouteJourneyTimeUp::compareObjects(Object *obj1, Object *obj2)
{
  Route *route1=(Route *)obj1;
  Route *route2=(Route *)obj2;

  int t1=route1->calc_journeyTime();
  int t2=route2->calc_journeyTime();

  t1=t1*1000+route1->get_numConnections();
  t2=t2*1000+route2->get_numConnections();
  if(t1<t2) return -1;
  if(t1>t2) return 1;
  return 0;
}

int SortRouteConnectionsUp::compareObjects(Object *obj1, Object *obj2)
{
  Route *route1=(Route *)obj1;
  Route *route2=(Route *)obj2;

  int t1=route1->get_numConnections();
  int t2=route2->get_numConnections();

  t1=t1*1000+route1->calc_journeyTime();
  t2=t2*1000+route2->calc_journeyTime();
  if(t1<t2) return -1;
  if(t1>t2) return 1;
  return 0;
}

int SortRouteWeightTime::compareObjects(Object *obj1, Object *obj2)
{
  Route *route1=(Route *)obj1;
  Route *route2=(Route *)obj2;



  int t1, t2;
  int c1, c2;
  int res1, res2;

  t1=(route1->calc_journeyTime()*10)*80/100;
  t2=(route2->calc_journeyTime()*10)*80/100;

  c1=(route1->get_numConnections()*100)*20/100;
  c2=(route2->get_numConnections()*100)*20/100;

  res1=t1+c1;
  res2=t2+c2;

  if(res1<res2) return -1;
  if(res1>res2) return 1;
  return 0;
}

/*	==============================================
	Output operators
	============================================== */


ostream& operator<<(ostream& os, Route& s)
{
  os << "Connections:  " << s.get_numConnections() << endl;
  os << "Journey time: " << s.calc_journeyTime() << endl << endl;
  
  ObjectListIterator it;
  Connection *conn=s.first_connection(it);
  while(conn) {
	os << conn;
	conn=s.next_connection(it);
  }
  return os;
}

ostream& operator<<(ostream& os, Route* s)
{
  os << *s;
  return os;
}


ostream& operator<<(ostream& os, Connection& s)
{
  Station *fromStation=s.get_fromStation();
  Station *toStation=s.get_toStation();
  Line *line=s.get_useLine();
  
  os << "(" << s.get_timeToChange() << ")";
  os << "  From " << fromStation->get_name() << " to " << toStation->get_name();
  if(s.get_byFeet()) {
	os << " by feet ";
  } else {
	os << " line " << line->get_name() << " [" << s.get_directionName() << "] ";
  } 
  os << s.calc_time() << " min.";
  os << endl;

  // passed stations
  Station *passedStation=s.first_passedStation();
  while(passedStation) {
	os << "     - " << passedStation->get_name() << endl;
	passedStation=s.next_passedStation();
  }
  return os;
}

ostream& operator<<(ostream& os, Connection *s)
{
  os << *s;
  return os;
}

