/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/



#include "SMXMLSubwayMapParser.h"


// =======================================================================
// general methods of subway map parser
// =======================================================================

smerr_t SMXMLSubwayMapParser::parse(const char *filename, SubwayMap *subwayMapRoot)
{
  int rc;

  // reset state
  
  _state=smap_none;

  _subwayMapRoot=subwayMapRoot;

  // initialize parser

  rc=initParser();
  if(rc) return rc;

  // start parsing the document
  rc=parseDocument(filename);
  if(rc) return rc;

  // finished parsing. add serving line information to each station
  _subwayMapRoot->add_servingLinesToStations();
  
  return 0;
}

// ----------------------------------------------------------------------

smerr_t SMXMLSubwayMapParser::parse_element(XMLElement& e)
{
  // if(_dumpMode>=dump_min)
  //	cinfo << "parse(" << e.get_name() << ")" << endl;

  // ----------------------------------
  // start document
  // ----------------------------------

  if(e.get_name()==tag_subway_map) {
	if(_state!=smap_none) {

	  return exception(new XMLException((const String&)"invalid tag, already in <subway_map", 
									smeXML_INVALID_TAG,	e.get_name(), _line, _column));

	}
	return start_subway_map(e);
  }

  if(e.get_name()==tag_info) {
	return parse_info(e);
  }



  // -----------------------------------
  // parse stations
  // -----------------------------------
  if(e.get_name()==tag_stations) { 
	if(_state!=smap_start) { // invalid state
	  return exception(new XMLException((const String&)"invalid element for <subway_map>", 
									smeXML_INVALID_TAG,	e.get_name(), _line, _column));
	}
	_state=smap_stations;
	//cout << "_state=smap_stations" << endl;
	return smeOK;
  }

  if(e.get_name()==tag_station) {
	return start_station(e);
  }

  if(e.get_name()==tag_serving_line || e.get_name()==tag_sl) {
	return parse_serving_line(e);
  }

  if(e.get_name()==tag_pos) {
  	// <pos lat="latitude" lon="longitude"/>
  	// Is allowed but will be ignored
  	return smeOK;
  }


  // ------------------------------------
  // parse lines
  // ------------------------------------
  if(e.get_name()==tag_lines) {
	if(_state!=smap_start) { // invalid state
	  return exception(new XMLException((const String&)"invalid element for <subway_map>", 
									smeXML_INVALID_TAG,	e.get_name(), _line, _column));

	}
	_state=smap_lines;
	// cout << "_state=smap_lines" << endl;
	return smeOK;
  }

  if(e.get_name()==tag_line) {
	return start_line(e);
  }
  if(e.get_name()==tag_route) {
	return start_route(e);
  }
  if(e.get_name()==tag_stop) {
	return start_stop(e);
  }
  if(e.get_name()==tag_stop_ctl) {
	return parse_ctl(e);
  }
  if(e.get_name()==tag_stop_cts) {
	return parse_cts(e);
  }
  if(e.get_name()==tag_optime) {
	return parse_optime(e);
  }

  if(_dumpMode>dump_none) 
	cerror << "Unknown element: " << e.get_name() << endl;

  if(_state==smap_none) {
	// invalid document
	return exception(new XMLException((const String&)"invalid document element",
								  smeXML_INVALID_DOCUMENT, e.get_name(), _line, _column));
  }

  return exception(new XMLException((const String&)"unknown element", smeXML_UNKNOWN_TAG,
								e.get_name(), _line, _column));
}

// ----------------------------------------------------------------------

smerr_t SMXMLSubwayMapParser::finish_element(XMLElement& e)
{
  // if(_dumpMode>=dump_min)
  //	cinfo << "finish(" << e.get_name() << ")" << endl;

  if(e.get_name()==tag_subway_map) {
	// no more tags allowed
	_state=smap_finished;
  }

  if(e.get_name()==tag_station) {
	return finish_station(e);
  }

  if(e.get_name()==tag_line) {
	return finish_line(e);
  }

  if(e.get_name()==tag_route) {
	return finish_route(e);
  }
  if(e.get_name()==tag_stop) {
	return finish_stop(e);
  }

  if(e.get_name()==tag_stations ||
	 e.get_name()==tag_info ||
	 e.get_name()==tag_lines) {
	_state=smap_start;
  }

  // the cannot be an invalid finished element
  return smeOK;
}

// =======================================================================
// methods for parsing the information part of the subway map
// =======================================================================

smerr_t SMXMLSubwayMapParser::parse_info(XMLElement& e)
{
  if(_dumpMode>=dump_min)
	cinfo << "--- Info --- " << e.get_name() << endl;

  XMLAttribute& aname=e.get_attribute(attr_info_name);
  XMLAttribute& acity=e.get_attribute(attr_info_city);
  XMLAttribute& adescr=e.get_attribute(attr_info_descr);
  XMLAttribute& aversion=e.get_attribute(attr_info_version);
  XMLAttribute& acreated=e.get_attribute(attr_info_created);
  XMLAttribute& alastchange=e.get_attribute(attr_info_last_change);
  XMLAttribute& acreatedby=e.get_attribute(attr_info_created_by);

  _subwayMapRoot->set_name(aname.get_value());
  _subwayMapRoot->set_city(acity.get_value());
  _subwayMapRoot->set_descr(adescr.get_value());
  _subwayMapRoot->set_version(aversion.get_value());
  _subwayMapRoot->set_created(acreated.get_value());
  _subwayMapRoot->set_last_change(alastchange.get_value());
  _subwayMapRoot->set_created_by(acreatedby.get_value());

  if(_dumpMode==dump_all) {
	cinfo << "name=" << aname.get_value() << endl
		  << "city=" << acity.get_value() << endl
		  << "descr=" << adescr.get_value() << endl
		  << "version=" << aversion.get_value() << endl
		  << "created=" << acreated.get_value() << endl
		  << "last change=" << alastchange.get_value() << endl
		  << "created by=" << acreatedby.get_value() << endl;
  }

  return smeOK;
}

// =======================================================================
// methods for parsing stations
// =======================================================================

smerr_t SMXMLSubwayMapParser::start_station(XMLElement& e)
{
  if(_dumpMode>=dump_min) 
	cinfo << "--- Station ---" << endl;

  if(_state!=smap_stations) {

	return exception(new XMLException((const String&)"invalid element: <station> outside <stations>", 
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));

  }

  XMLAttribute& aname=e.get_attribute(attr_station_name);
  XMLAttribute& aid=e.get_attribute(attr_station_id);

  if(aname==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: name",
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  if(aid==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: id",
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }


  // Create new station
  if((_actStation=new Station(aname.get_value(), aid.get_value()))==NULL) {
	return exception(new MemoryException());
  }


  if(_dumpMode==dump_all) {
	cinfo << "name=" << aname.get_value() << endl;
  }

  _state=smap_in_station;
  return smeOK;
}

// ----------------------------------------------------------------------

smerr_t SMXMLSubwayMapParser::parse_serving_line(XMLElement& e)
{
	
	
#if 0	

	// Serving Line Tags <sl> will be ignored


  if(_dumpMode>=dump_min) {
	cinfo << "  --> Serving Line" << endl;
  }


  if(_state!=smap_in_station) {
	return exception(new XMLException((const String&)"invalid element: <sl> outside <station>", 
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));
  }

  if(!_actStation) {
	return exception(new XMLException((const String&)"Element <sl> but no <station>", smeXML_INTERNAL_ERROR,
								  e.get_name(), _line, _column));
  }

  XMLAttribute& aid=e.get_attribute(attr_serving_line_id);

  if(aid==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: id", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  ServingLine *servingLine=new ServingLine;
  servingLine->set_id(aid.get_value());
  _actStation->add_servingLine(servingLine);

  if(_dumpMode>=dump_min) {
	cinfo << "      " << aid.get_value() << endl;
  }
#endif

  return smeOK;
}

// ----------------------------------------------------------------------

smerr_t SMXMLSubwayMapParser::finish_station(XMLElement& e)
{
  if(_dumpMode>=dump_min) {
	cinfo << "--- Station finished ---" << endl << endl;
  }

  if(_state!=smap_in_station) {
	return exception(new XMLException((const String&)"interal error: _state!=smap_in_station",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }

  // Add station to station list
  _subwayMapRoot->add_station(_actStation);
  
  _actStation=NULL;

  _state=smap_stations;
  return smeOK;
}

// =======================================================================
// methods for parsing lines
// =======================================================================

smerr_t SMXMLSubwayMapParser::start_line(XMLElement& e)
{
  if(_dumpMode>=dump_min)
	cinfo << "--- Line ---" << endl;

  if(_state!=smap_lines) {
	return exception(new XMLException((const String&)"invalid element: <line> outside <lines>", 
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));
  }

  _state=smap_in_line;

  XMLAttribute& aname=e.get_attribute(attr_line_name);
  XMLAttribute& aid=e.get_attribute(attr_line_id);
  XMLAttribute& aclass=e.get_attribute(attr_line_class);

  if(aid==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: id", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }
  if(aname==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: name", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  if(_dumpMode>=dump_min) {
	cinfo << "name=" << aname.get_value() << endl
		  << "id=" << aid.get_value() << endl
		  << "class=" << aid.get_value() << endl;
  }
  if((_actLine=new Line(aname.get_value(), aid.get_value(), aclass.get_value()))==NULL) {
	return exception(new MemoryException());
  }

  
  return smeOK;
}

smerr_t SMXMLSubwayMapParser::finish_line(XMLElement& e)
{
  if(_dumpMode>dump_none)
	cinfo << "--- Line finished ---" << endl;

  if(_state!=smap_in_line) {

	return exception(new XMLException((const String&)"internal error: _state!=smap_in_line",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }
  if(!_actLine) {

	return exception(new XMLException((const String&)"internal error: no Line object allocated",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }

  _subwayMapRoot->add_line(_actLine);
  _actLine=NULL;

  _state=smap_lines;

  return smeOK;
};

// ----------------------------------------------------------------------

smerr_t SMXMLSubwayMapParser::start_route(XMLElement& e)
{
  if(_dumpMode>=dump_min)
	cinfo << "-- Route of Line --" << endl;

  if(_state!=smap_in_line) {

	return exception(new XMLException((const String&)"invalid element: <route> outside <line>",
								  smeXML_INVALID_TAG,  e.get_name(), _line, _column));
  }

  XMLAttribute& aname=e.get_attribute(attr_route_name);

  // create new route object
  if((_actRouteOfLine=new RouteOfLine(aname.get_value()))==NULL) {
	return smeOUT_OF_MEMORY;
  }

  XMLAttribute& aservstart=e.get_attribute(attr_route_service_start);
  XMLAttribute& aservend=e.get_attribute(attr_route_service_end);
  XMLAttribute& astartstation=e.get_attribute(attr_route_startstation);
  XMLAttribute& aendstation=e.get_attribute(attr_route_endstation);
  XMLAttribute& aoneway=e.get_attribute(attr_route_oneway);

  if(astartstation==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: startstation", 
								  smeXML_MISSING_ATTR,  e.get_name(), _line, _column));
  }
  if(aendstation==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: endstation", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  _actRouteOfLine->set_startstation(astartstation.get_value());
  _actRouteOfLine->set_endstation(aendstation.get_value());

  if(aservstart!=unknownAttribute)
	_actRouteOfLine->set_service_start(aservstart.get_value());
  if(aservend!=unknownAttribute)
	_actRouteOfLine->set_service_end(aservend.get_value());

  if(aoneway==attr_value_true) {
	_actRouteOfLine->set_oneway(1);
  }

  if(_dumpMode>=dump_min) {
	cinfo << "startstation=" << astartstation.get_value() << endl
		  << "endstation=" << aendstation.get_value() << endl
		  << "service_start" << aservstart.get_value() << endl
		  << "service_end" << aservend.get_value() << endl;
  }
  _state=smap_in_route;

  return smeOK;
}


smerr_t SMXMLSubwayMapParser::finish_route(XMLElement& e)
{
  if(_dumpMode>=dump_min)
	cinfo << "--- Route of line finished ---" << endl;

  if(_state!=smap_in_route) {
	return exception(new XMLException((const String&)"interal error: state!=smap_in_route",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }

  if(!_actRouteOfLine) {

	return exception(new XMLException((const String&)"internal error: no _actRouteOfLine object allocated",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }

  if(!_actLine) {

	return exception(new XMLException((const String&)"internal error: no _actLine object allocated",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }
  
  _actRouteOfLine->set_line(_actLine);
  _actLine->add_routeOfLine(_actRouteOfLine);
  _actRouteOfLine=NULL;

  _state=smap_in_line;
  
  return smeOK;
}

// ----------------------------------------------------------------------

smerr_t SMXMLSubwayMapParser::start_stop(XMLElement& e)
{
  if(_dumpMode>=dump_min) 
	cinfo << "  ---> Stop at" << endl;

  if(_state!=smap_in_route) {

	return exception(new XMLException((const String&)"invalid element: <stop> outside <route>",
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));
  }

  if(!_actRouteOfLine) {

	return exception(new XMLException((const String&)"internal error: no _actRouteOfLine object allocated",
								  smeXML_INTERNAL_ERROR, e.get_name(), _line, _column));
  }

  XMLAttribute& aat=e.get_attribute(attr_stop_at);
  XMLAttribute& attn=e.get_attribute(attr_stop_ttn);
  XMLAttribute& attp=e.get_attribute(attr_stop_ttp);

  if(aat==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: at", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }
  if(attn==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: ttn", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }
 

  // Find station by ID
  Station *station=_subwayMapRoot->find_stationById(aat.get_value());
  if(!station) {
	return exception(new XMLException((const String&)"invalid station reference", smeXML_INVALID_STATION_REF,
								  e.get_name(), _line, _column));
  }

  int ttn=attn.get_value().to_int();
  int ttp=attp.get_value().to_int();

  if(_dumpMode>dump_min) {
	cinfo << "      id=" << aat.get_value() << endl
		  << "      ttn=" << ttn << endl
		  << "      ttp=" << ttp << endl;
  }

  // create new stop object
  if((_actStopAt=new StopAt(station, ttn, ttp))==NULL) {
	return exception(new MemoryException());
  }

  _state=smap_in_stop;

  return smeOK;
}

smerr_t SMXMLSubwayMapParser::finish_stop(XMLElement& e)
{
  if(_dumpMode>=dump_min) {
	cinfo << endl;
  }
  _actRouteOfLine->add_stop(_actStopAt);
  _actStopAt=NULL;

  _state=smap_in_route;

  return smeOK;
}

smerr_t SMXMLSubwayMapParser::parse_ctl(XMLElement& e)
{
  if(_dumpMode>=dump_min) {
	cinfo << "    ---> parse ctl ---" << endl;
  }

  if(_state!=smap_in_stop) {
	return exception(new XMLException((const String&)"invalid element: <ctl> outside <stop>",
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));
  }

  XMLAttribute& aid=e.get_attribute(attr_ctl_id);
  String& ctlid=aid.get_value();

  if(aid==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: id", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  XMLAttribute& attc=e.get_attribute(attr_ctl_ttc);
  int ttc=attc.get_value().to_int();

  if(_dumpMode>dump_min)
	cinfo << "         " << ctlid << endl;

  // Set direction attribute
  eDirection fromdir=both;
  XMLAttribute& afromdir=e.get_attribute(attr_ctl_from_dir);
  String& sfromdir=afromdir.get_value();
  if(sfromdir==attr_value_forward) 
	fromdir=forward;
  if(sfromdir==attr_value_backwards)
	fromdir=backwards;

  eDirection todir=both;
  XMLAttribute& atodir=e.get_attribute(attr_ctl_to_dir);
  String& stodir=atodir.get_value();
  if(stodir==attr_value_forward)
	todir=forward;
  if(stodir==attr_value_backwards)
	todir=backwards;

  if(_dumpMode>dump_min)
	cinfo << "         " << ctlid << " from_dir=" << sfromdir << " to_dir=" << stodir << " ttc=" << ttc << endl;


  ChangeToLine *ctl=new ChangeToLine(ctlid, fromdir, todir, ttc);
  _actStopAt->add_changeToLine(ctl);
  _actStopAt->set_change(1);

  return smeOK;
}

smerr_t SMXMLSubwayMapParser::parse_cts(XMLElement& e)
{
  if(_dumpMode>=dump_min) {
	cinfo << "    ---> parse cts ---" << endl;
  }

  if(_state!=smap_in_stop) {
	return exception(new XMLException((const String&)"invalid element: <cts> outside <stop>",
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));
  }

  XMLAttribute& aid=e.get_attribute(attr_cts_id);
  String& ctsid=aid.get_value();

  if(aid==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: id", 
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  XMLAttribute& attc=e.get_attribute(attr_cts_ttc);
  int ttc=attc.get_value().to_int();

  if(_dumpMode>dump_min)
	cinfo << "         " << ctsid << endl;

  // Set direction attribute
  eDirection fromdir=both;
  XMLAttribute& afromdir=e.get_attribute(attr_cts_from_dir);
  String& sfromdir=afromdir.get_value();
  if(sfromdir==attr_value_forward) 
	fromdir=forward;
  if(sfromdir==attr_value_backwards)
	fromdir=backwards;

  eDirection todir=both;
  XMLAttribute& atodir=e.get_attribute(attr_cts_to_dir);
  String& stodir=atodir.get_value();
  if(stodir==attr_value_forward) 
	todir=forward;
  if(stodir==attr_value_backwards)
	todir=backwards;



  if(_dumpMode>dump_min)
	cinfo << "         " << ctsid << " from_dir=" << sfromdir << " to_dir=" << stodir << " ttc=" << ttc << endl;


  ChangeToStation *cts=new ChangeToStation(ctsid, fromdir, todir, ttc);
  _actStopAt->add_changeToStation(cts);
  _actStopAt->set_change(1);

  return smeOK;
}

smerr_t SMXMLSubwayMapParser::parse_optime(XMLElement& e)
{
  if(_dumpMode>=dump_min) {
	cinfo << "    ---> parse optime ---" << endl;
  }

  if(_state!=smap_in_route) {

	return exception(new XMLException((const String&)"invalid element: <optime> outside <route>",
								  smeXML_INVALID_TAG, e.get_name(), _line, _column));
  }

  /* ----------------------------------------------------
	 Format:

	 XML: <optime days="<days>" hours="<hours>"/>

	 Examples:

	 days="1234567": Monday to Sunday (1=Mon, 7=Sun)
	 days="12345":   Mondey to Friday (1=Mon, 5=Fri)
	 days="67":      Saturday and Sunday (6=Sat, 7=Sun)

	 hours="6":      06:00
	 hours="6,7,8":  06:00, 07:00, 08:00
	 hours="0-2":    00:00 to 02:00
	 hours="5-23":   05:00 to 23:00
	 ---------------------------------------------------- */

  XMLAttribute& days=e.get_attribute(attr_optime_days);
  XMLAttribute& hours=e.get_attribute(attr_optime_hours);

  if(days==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: days",
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }
  if(hours==unknownAttribute) {
	return exception(new XMLException((const String&)"missing attribute: hours",
								  smeXML_MISSING_ATTR, e.get_name(), _line, _column));
  }

  String& sdays=days.get_value();
  String& shours=hours.get_value();

  cout << "attr: days=" << (const char *)sdays << ", hours=" << (const char *)shours << endl;

  unsigned int i;
  int numdays=0;
  int idays[7];
  int numhours=0;
  unsigned int ihours[24];
  int t=0;

  // decode days
  for(i=0; i<sdays.length(); i++) {
	char c=sdays[i];
	if(c<'1' || c>'7') { // only 1..7
	  cerror << "invalid day" << c << endl;
	  continue;
	}
	idays[numdays++]=c-0x30; // make integer from character
  }

  // decode hours
  int fromhour=-1, tohour=-1;
  char tmpstr[30];

  for(i=0; i<=shours.length(); i++) { // incl. 0-byte
	char c=shours[i];

	if(c==',' || c==0) {
	  int value=atoi(tmpstr);
	  t=0;
	  memset(tmpstr, 0, 30);

	  if(fromhour>=0) {
		// expression was like "0-4"
		tohour=value;
		for(int h=fromhour; h<=tohour; h++)
		  ihours[numhours++]=h;
	  } else {
		// expression was like ",5,"
		ihours[numhours++]=value;
	  }
	  fromhour=-1;
	  tohour=-1;
	  continue;
	}

	if(c=='-') {
	  // expression was like "5-"
	  int value=atoi(tmpstr);
	  t=0;
	  memset(tmpstr, 0, 30);
	  fromhour=value;
	  continue;
	}
	tmpstr[t++]=c;
  }

  OperationTime& optime=_actRouteOfLine->get_optime();

  // Set operation times from string
  for(int d=0; d<numdays; d++) {
	// cast int value to EDay => EDay starts a 0, iday starts at 1
	OperationTime::EDay eday=static_cast<OperationTime::EDay>(idays[d]-1);

	if(_dumpMode>=dump_min) {
		cinfo << "day=" << eday << ": ";
		for(i=0; i<numhours; i++) {
		  cinfo << ihours[i] << " ";
		}
		cinfo << endl;
	}
	optime.set_ophours(eday, ihours, numhours, true);
  }
  return smeOK;
}



smerr_t SMXMLSubwayMapParser::start_subway_map(XMLElement& e)
{
	if(_dumpMode>=dump_min)
		cinfo << "--- subway_map --- " << e.get_name() << endl;

	_state=smap_start;
    	
	XMLAttribute& aversion=e.get_attribute(attr_subway_map_version);
	if(aversion==unknownAttribute) {
		_version=1;
		cwarning << "no version tag, assuming version=1" << endl;
	} else {
		XMLString& version=aversion.get_value();
		_version=version.to_int();
		if(_dumpMode>=dump_min)
			cinfo << "version=" << _version << endl;
	}
	return smeOK;  
}


smerr_t SMXMLSubwayMapParser::finsh_subway_map(XMLElement& e)
{
	if(_dumpMode>=dump_min) {
		cinfo << "--- subway_map ends ---" << endl;
	}
	_state=smap_finished;
}