/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#ifndef __SMXMLSUBWAYMAPPARSER_H__
#define __SMXMLSUBWAYMAPPARSER_H__


#include "baseclasses.h"
#include "XMLBase.h"

#include "SMError.h"
#include "Exceptions.h"

#include "SubwayMap.h"
#include "debug.h"

/*	===========================================
	Defines for SubwayMap XML Tags
	=========================================== */

// ---- XML Tag <subway_map>
const XMLString tag_subway_map		= "subway_map";
const XMLString attr_subway_map_version = "version";

// XML Tag <info>
const XMLString tag_info			= "info";
const XMLString attr_info_name		= "name";
const XMLString attr_info_city		= "city";
const XMLString attr_info_descr		= "descr";
const XMLString attr_info_version	= "version";
const XMLString attr_info_created	= "created";
const XMLString attr_info_last_change = "last_change";
const XMLString attr_info_created_by= "created_by";

// ---- XML Tag <lines>
const XMLString tag_lines			= "lines";

const XMLString tag_line			= "line";
const XMLString attr_line_name		= "name";
const XMLString attr_line_color		= "color";
const XMLString attr_line_id		= "id";
const XMLString attr_line_class		= "class";

const XMLString tag_route			= "route";
const XMLString attr_route_name		= "name";
const XMLString attr_route_service_start = "service_start";
const XMLString attr_route_service_end = "service_end";
const XMLString attr_route_startstation= "startstation";
const XMLString attr_route_endstation  ="endstation";
const XMLString attr_route_oneway   = "oneway";

const XMLString tag_stop			= "stop";
const XMLString attr_stop_at		= "at";
const XMLString attr_stop_ttn		= "ttn";
const XMLString attr_stop_ttp		= "ttp";
const XMLString tag_alt_route		= "alt_route";

const XMLString tag_stop_ctl		= "ctl";
const XMLString attr_ctl_id			= "id";
const XMLString attr_ctl_from_dir	= "from_dir";
const XMLString attr_ctl_to_dir		= "to_dir";
const XMLString attr_ctl_ttc		= "ttc";

const XMLString tag_stop_cts		= "cts";
const XMLString attr_cts_id			= "id";
const XMLString attr_cts_from_dir	= "from_dir";
const XMLString attr_cts_to_dir		= "to_dir";
const XMLString attr_cts_ttc		= "ttc";

const XMLString tag_optime			= "optime";
const XMLString attr_optime_days	= "days";
const XMLString attr_optime_hours	= "hours";

// ---- XML Tag <stations>
const XMLString tag_stations		= "stations";

const XMLString tag_station			= "station";
const XMLString attr_station_name	= "name";
const XMLString attr_station_id		= "id";

const XMLString tag_serving_line	= "serving_line";
const XMLString tag_sl				= "sl";
const XMLString attr_serving_line_id= "id";
const XMLString tag_pos				= "pos";


// ---- Div. XML values and stuff
const XMLString attr_value_true		= "true";
const XMLString attr_value_false	= "false";
const XMLString attr_value_forward	= "forw";
const XMLString attr_value_backwards= "backw";

/**
 * Baseclasses of the Subway Map XML Parser.
 * This class has common methods to store subwaymap relevant elements into a subway map network.
 * It doesn't have methods for parsing XML code. These methods will be provided by each XMLSubwayMapParser child class (eg. XMLSubwayMapParser_Qt). 
 */


class SubwayMap;

class SMXMLSubwayMapParser : public Object
{
 private:
  enum eParserState {
	smap_none=0,
	smap_start,			// got element <subway_map>
	smap_info,			// got element <info>
	smap_stations,		// got element <stations>
	smap_in_station,	// got element <station>
	smap_lines,			// got element <lines>
	smap_in_line,		// got element <line>
	smap_in_route,		// got element <route>
	smap_in_stop,		// got element <stop>
	smap_finished		// <subway_map> closed
  };


 public:
  virtual ~SMXMLSubwayMapParser() {
	
  };

  virtual smerr_t parse(const char *filename, SubwayMap *subwayMapRoot);

  SubwayMap *get_subwayMapRoot() { return _subwayMapRoot; }


  int get_line() { return _line; }
  int get_column() { return _column; }

 protected:
  SMXMLSubwayMapParser() : _exception(NULL), _subwayMapRoot(NULL), _state(smap_none), _version(0),
	_dumpMode(dump_none), _actStation(NULL), _actLine(NULL), _actRouteOfLine(NULL), 
	_actStopAt(NULL) {} 

  // -----------------------------------------------------------
  // abstract methods, that must be implemented by each subclass
  // -----------------------------------------------------------
  virtual smerr_t initParser()=0;
  virtual smerr_t parseDocument(const char *filename)=0;

  // -----------------------------------------------------------
  // interface for parsing an element
  // This method will be called by the subclass's element
  // parsing method.
  // -----------------------------------------------------------
  smerr_t parse_element(XMLElement& e);
  smerr_t finish_element(XMLElement& e);

  // -----------------------------------------------------------
  // element dependent parsing methods
  // -----------------------------------------------------------

  smerr_t start_subway_map(XMLElement& e);
  smerr_t finsh_subway_map(XMLElement& e);

  smerr_t parse_info(XMLElement& e);			// info 

  // parse stations
  smerr_t start_station(XMLElement& e);	 	// parse a station
  smerr_t parse_serving_line(XMLElement& e); // parse serving line tag (in station)
  smerr_t finish_station(XMLElement& e);	 	// end of station tag

  // parse lines
  smerr_t start_line(XMLElement& e);			// parse a line
  smerr_t start_route(XMLElement& e);		// parse route aof line
  smerr_t start_stop(XMLElement& e);		// parse one stop
  smerr_t parse_ctl(XMLElement& e);			// parse one ctl element
  smerr_t parse_cts(XMLElement& e);			// parse one cts element
  smerr_t parse_optime(XMLElement& e);		// parse one optime element
  smerr_t finish_stop(XMLElement& e);		// finish one stop
  smerr_t finish_route(XMLElement& e);		// finish route
  smerr_t finish_line(XMLElement& e);		// end of line tag

 protected:
  // -----------------------------------
  // Error handling/Exception simulation
  // -----------------------------------
  smerr_t exception(const ExceptionBase *e) { 
	_exception=const_cast<ExceptionBase*>(e); 
	return _exception->get_extype();
  }

  ExceptionBase *_exception;

 public:
  ExceptionBase *get_exception() { return _exception; }
  void clean_exception() {
	if(_exception) delete _exception;
	_exception=NULL;
  }

 protected:
  SubwayMap *_subwayMapRoot;
  eParserState _state;
  int _version;

  void set_linecolumn(int line, int column) {
	_line=line;
	_column=column;
  }

  int _line;
  int _column;

 public:
  // -----------------------------------------------------------
  // dump document contents while parsing the document
  // -----------------------------------------------------------

  enum eDumpMode {
	dump_none=0,		// no output during parsing
	dump_errors,		// dump errors to std error
	dump_min,			// minimal output (tags tags)
	dump_all			// dump tags and attributes
  };

  void set_dumpMode(eDumpMode mode) {
	_dumpMode=mode;
  }
 protected:
  eDumpMode _dumpMode;


 private:
  // -----------------------------------------------------------
  // temporary objects
  // -----------------------------------------------------------
  Station *_actStation;			// Actual station
  Line *_actLine;				// Actual line
  RouteOfLine *_actRouteOfLine;	// Actual route of line
  StopAt *_actStopAt;			// Actual StopAt
};



#endif
