/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#include "debug.h"
#include "SubwayGraph.h"



SubwayGraph::~SubwayGraph()
{
  if(_nodeArray) {
	delete _nodeArray;
  }
}


smerr_t SubwayGraph::build()
{
  smerr_t rc;

  if(!_subwayMap) {
	cerror << "SubwayGraph::build() no subway map" << endl;
	return smeGRAPH_NO_SUBWAYMAP;
  }


  cout << "Day=" << static_cast<unsigned int>(_opDay) << " Hour=" << _opHour << endl;

  if((rc=create_nodes())!=smeOK) {
	cerror << "error creating nodes rc=" << rc << endl;
	return rc;
  }

  if((rc=create_edges())!=smeOK) {
	cerror << "error creationg edges rc=" << rc << endl;
	return rc;
  }

  return smeOK;
}

smerr_t SubwayGraph::reset(SGSettings settings)
{
  for(int index=0; index<get_numNodes()-2; index++) {
	SGNode *node=(SGNode *)get_nodeAt(index);

	if(!node)
	  continue;

	// reset distance to start and previous node
	node->set_distance(INFINITE);
	node->set_prevNode(NULL);

	// 
	SGEdge *edge=node->first_edge();

	while(edge) {
	  if(edge->get_type()==SGEdge::changeToLine || 
		 edge->get_type()==SGEdge::changeToStation) {

		edge->set_weight(edge->get_realweight());

		if(settings==direct) {
		  edge->set_weight(edge->get_weight()*10);
		}

		if(settings==dontwalk) {
		  if(edge->get_type()==SGEdge::changeToStation)
			edge->set_weight(edge->get_weight()*10);
		}
	  }
	  edge=node->next_edge();
	}
  }
  return smeOK;
}

smerr_t SubwayGraph::cleanup()
{
  return smeOK;
}


smerr_t SubwayGraph::create_nodes()
{
  if(!_subwayMap) {
	return smeGRAPH_NO_SUBWAYMAP;
  }

  // get the number of nodes

  int numnodes=_subwayMap->get_numStationsOnAllLines();
  numnodes+=2;	// add nodes for start and destination
  cdebug << "numnodes=" << numnodes << endl;


  if(_nodeArray) {
	delete _nodeArray;
	_nodeArray=NULL;
  }

  if((_nodeArray=new ObjectArray(numnodes))==NULL) {
	cerror << "Error creating _nodeArray" << endl;
	return smeOUT_OF_MEMORY;
  }

  int index=0;

  ObjectListIterator lineIt;
  Line *line=_subwayMap->first_line(lineIt);
  while(line) {

	// use line only if it operates now...
	/*
	if(!line->operates_at(_opDay, _opHour)) {
	  line=_subwayMap->next_line(lineIt);
	  continue;
	}
	*/
	
	ObjectListIterator routeIt;
	RouteOfLine *routeOfLine=line->first_routeOfLine(routeIt);
	while(routeOfLine) {

	  // use route of line only if it operates now
	  /*
	  if(!routeOfLine->operates_at(_opDay, _opHour)) {
		routeOfLine=line->next_routeOfLine(routeIt);
		continue;
	  }
	  */

	  ObjectListIterator stopIt;
	  StopAt *stopAt=routeOfLine->first_stop(stopIt);
	  while(stopAt) {

		// create new node and add it to node array
		SGNode *node=new SGNode(stopAt->get_station(), routeOfLine);
		if(!node) {
		  cerror << "error creating node" << endl;
		  return smeOUT_OF_MEMORY;
		}

		_nodeArray->set_at(index++, node);

		stopAt=routeOfLine->next_stop(stopIt);
	  }

	  routeOfLine=line->next_routeOfLine(routeIt);
	}

	line=_subwayMap->next_line(lineIt);
  }
  
  return smeOK;
}


smerr_t SubwayGraph::create_edges()
{
  smerr_t rc;

  if(!_subwayMap) {
	return smeGRAPH_NO_SUBWAYMAP;
  }

  if(!_nodeArray) {
	return smeGRAPH_NO_NODES;
  }

  ObjectListIterator lineIt;
  Line *line=_subwayMap->first_line(lineIt);
  while(line) {

	// use line only if it operates now
	if(!line->operates_at(_opDay, _opHour)) {
	  line=_subwayMap->next_line(lineIt);
	  continue;
	}
	

	ObjectListIterator routeIt;
	RouteOfLine *routeOfLine=line->first_routeOfLine(routeIt);
	while(routeOfLine) {

	  // use route only if it operates now
	  if(!routeOfLine->operates_at(_opDay, _opHour)) {
		routeOfLine=line->next_routeOfLine(routeIt);
		continue;
	  }


	  if((rc=process_routeOfLine(routeOfLine))!=smeOK) {
		cerror << "error processing routeOfLine rc=" << rc << endl;
		return rc;
	  }
	  routeOfLine=line->next_routeOfLine(routeIt);
	}
	
	line=_subwayMap->next_line(lineIt);
  }

  return smeOK;
}

smerr_t SubwayGraph::process_routeOfLine(RouteOfLine *routeOfLine)
{
  smerr_t rc;

  ObjectListIterator stopIt;
  StopAt *stop=routeOfLine->first_stop(stopIt);
  while(stop) {
	
	Station *station=stop->get_station();
	SGNode *node=find_nodeWithStationOnRoute(station, routeOfLine);
	if(!node) {
	  cerror << "process_routeOfLine(): node not found" << endl;
	  return smeGRAPH_NODE_NOT_FOUND;
	}

	// --------------------------------
	// -- edge to next node on route --
	ObjectListIterator nextStopIt=stopIt;
	StopAt *nextStop=routeOfLine->next_stop(nextStopIt);
	if(nextStop) {
	  Station *nextStation=nextStop->get_station();
	  SGNode *nextNode=find_nodeWithStationOnRoute(nextStation, routeOfLine);
	  if(!nextNode) {
		cerror << "process_routeOfLine: node not found (next)" << endl;
		return smeGRAPH_NODE_NOT_FOUND;
	  }

	  SGEdge *edge=new SGEdge(stop->get_timeToNext(), nextNode);
	  node->add_edge(edge);
	}

	// ------------------------------------
	// -- edge to previous node on route --
	if(!routeOfLine->get_oneway()) {
	  ObjectListIterator prevStopIt=stopIt;
	  StopAt *prevStop=routeOfLine->prev_stop(prevStopIt);
	  if(prevStop) {
		Station *prevStation=prevStop->get_station();
		SGNode *prevNode=find_nodeWithStationOnRoute(prevStation, routeOfLine);
		if(!prevNode) {
		  cerror << "process_routeOfLine: node not found (prev)" << endl;
		  return smeGRAPH_NODE_NOT_FOUND;
		}

		SGEdge *edge=new SGEdge(stop->get_timeToPrev(), prevNode);
		node->add_edge(edge);
	  }
	}

	// -----------------------------------------
	// --  process change to line from node   --

	ObjectListIterator ctlIt;
	ChangeToLine *ctl=stop->first_changeToLine(ctlIt);
	while(ctl) {

	  if((rc=add_ctlEdges(node, ctl))!=smeOK) {
		return rc;
	  }

	  ctl=stop->next_changeToLine(ctlIt);
	}

	// -----------------------------------------
	// -- process change to station from node --

	ObjectListIterator ctsIt;
	ChangeToStation *cts=stop->first_changeToStation(ctsIt);
	while(cts) {

	  if((rc=add_ctsEdges(node, cts))!=smeOK) {
		return rc;
	  }

	  cts=stop->next_changeToStation(ctsIt);
	}

	stop=routeOfLine->next_stop(stopIt);
  }

  return smeOK;
}

SGNode *SubwayGraph::find_nodeWithStationOnRoute(Station *station, RouteOfLine *routeOfLine)
{
  for(int i=0; i<_nodeArray->length()-2; i++) {
	SGNode *node=(SGNode *)_nodeArray->get_at(i);
	if(node->get_station()==station && node->get_routeOfLine()==routeOfLine) {
	  // node found
	  return node;
	}
  }
  return NULL;
}

smerr_t SubwayGraph::add_ctlEdges(SGNode *node, ChangeToLine *ctl)
{
  Station *station=node->get_station();
  String& lineid=ctl->get_id();

  if(!station || !lineid) {
	cerror << "station or line not found" << endl;
	return smeGRAPH_MISSING_PARAM;
  }

  for(int index=0; index<_nodeArray->length()-2; index++) {
	SGNode *target=(SGNode *)_nodeArray->get_at(index);
	if(!target)
	  continue;

	Station *targetStation=target->get_station();
	if(targetStation==station) {
	  Line *targetLine=target->get_line();

	  if(targetLine->get_id()==lineid && targetLine->operates_at(_opDay, _opHour) ) {
		SGEdge *edge=new SGEdge(ctl->get_timeToChange()*_ctlFactor, target, SGEdge::changeToLine);
		if(!edge) {
		  cerror << "out of memory" << endl;
		  return smeOUT_OF_MEMORY;
		}
		node->add_edge(edge);
	  }
	}
  }
  return smeOK;
}

smerr_t SubwayGraph::add_ctsEdges(SGNode *node, ChangeToStation *cts)
{
  String& stationId=cts->get_id();

  for(int index=0; index<_nodeArray->length()-2; index++) {
	SGNode *target=(SGNode *)_nodeArray->get_at(index);
	if(!target)
	  continue;

	Station *targetStation=target->get_station();
	if(targetStation->get_id()==stationId) {
	  SGEdge *edge=new SGEdge(cts->get_timeToChange()*_ctsFactor, target, SGEdge::changeToStation);
	  if(!edge) {
		cerror << "out of memory" << endl;
		return smeOUT_OF_MEMORY;
	  }
	  node->add_edge(edge);
	}
  }

  return smeOK;
}


int SubwayGraph::add_startNodeForStation(Station *station)
{
  if(!_nodeArray) {
	cerror << "no node array" << endl;
	return smeGRAPH_NO_NODES;
  }

  int startNodeIndex=_nodeArray->length()-2;

  // create node
  SGNode *node=new SGNode(station, SGNode::startNode);
  if(!node) {
	cerror << "cannot create start node" << endl;
	return smeOUT_OF_MEMORY;
  }

  // create edges to "real" stations

  for(int i=0; i<_nodeArray->length()-2; i++) {
	SGNode *targetNode=(SGNode *)_nodeArray->get_at(i);
	if(targetNode && targetNode->get_station()==station) {
	  SGEdge *edge=new SGEdge(0, targetNode, SGEdge::fromStartNode);
	  if(!edge) {
		cerror << "cannot create edge from start node" << endl;
		return smeOUT_OF_MEMORY;
	  }
	  node->add_edge(edge);
	}
  }
  _nodeArray->set_at(startNodeIndex, node);
  return startNodeIndex;
}

int SubwayGraph::add_destNodeForStation(Station *station)
{
  if(!_nodeArray) {
	cerror << "no node array" << endl;
	return smeGRAPH_NO_NODES;
  }

  int destNodeIndex=_nodeArray->length()-1;

  SGNode *node=new SGNode(station, SGNode::destNode);
  if(!node) {
	cerror << "cannot create dest node" << endl;
	return smeOUT_OF_MEMORY;
  }

  // find nodes with this station and add edges to destination
  for(int i=0; i<_nodeArray->length()-2; i++) {
	SGNode *sourceNode=(SGNode *)_nodeArray->get_at(i);
	if(sourceNode && sourceNode->get_station()==station) {
	  SGEdge *edge=new SGEdge(0, node, SGEdge::toDestNode);
	  if(!edge) {
		cerror << "cannot create edge to dest node" << endl;
		return smeOUT_OF_MEMORY;
	  }

	  sourceNode->add_edge(edge);
	}
  }
  _nodeArray->set_at(destNodeIndex, node);
  return destNodeIndex;
}

smerr_t SubwayGraph::remove_tempNodes()
{
  if(!_nodeArray) {
	return smeGRAPH_NO_NODES;
  }
  int startNodeIndex=_nodeArray->length()-2;
  int destNodeIndex=_nodeArray->length()-1;

  // remove start node and its edges
  SGNode *startNode=(SGNode *)_nodeArray->get_at(startNodeIndex);
  if(startNode) {
	_nodeArray->remove_at(startNodeIndex);
	delete startNode;
  }

  // remove dest node
  SGNode *destNode=(SGNode *)_nodeArray->get_at(destNodeIndex);
  if(destNode) {
	_nodeArray->remove_at(destNodeIndex);
	delete destNode;
  }

  // remove edges to dest node of all nodes
  for(int index=0; index<_nodeArray->length()-2; index++) {
	SGNode *node=(SGNode *)_nodeArray->get_at(index);
	if(node) {
	  node->remove_tempEdges();
	}
  }
  return smeOK;
}


bool SGNode::add_edge(SGEdge *edge)
{
  _edges.add((Object *)edge);

  return true;
}

bool SGNode::remove_tempEdges()
{
  ObjectListIterator it;
  SGEdge *edge=(SGEdge *)_edges.first(it);
  while(edge) {
	if(edge->get_type()==SGEdge::toDestNode) {
	  _edges.remove(edge);
	  delete edge;
	  edge=(SGEdge *)_edges.first(it);
	}
	edge=(SGEdge *)_edges.next(it);
  }
  return true;
}

void SGEdge::set_weight(unsigned int value, bool withfactor)
{
  _weight=value;

  if(withfactor) {
	if(_type==changeToLine) _weight=value*cCTLFACTOR;
	if(_type==changeToStation) _weight=value*cCTSFACTOR;
  }
}

SGEdge *SGNode::find_edgeToTarget(SGNode *target)
{
  ObjectListIterator it;
  SGEdge *edge=(SGEdge *)_edges.first(it);
  while(edge) {
	if(edge->get_target()==target) {
	  return edge;
	}
	edge=(SGEdge *)_edges.next(it);
  }
  return NULL;
}

// -- output methods
ostream& operator<<(ostream& os, SubwayGraph *g)
{
  for(int index=0; index<g->_nodeArray->length(); index++) {
	SGNode *node=(SGNode *)g->_nodeArray->get_at(index);
	if(node) {
	  os << "Node " << node;
	} else {
	  os << "NULL" << endl;
	}
  }
  os << endl;
  return os;
}

ostream& operator<<(ostream& os, SGNode *node)
{
  Station *station=node->get_station();
  Line *line=node->get_line();

  cout << "(" << station->get_name();

  if(node->get_type()==SGNode::startNode)
	cout << ")" << endl;
  else if(node->get_type()==SGNode::destNode)
	cout << ")" << endl;
  else
	cout << ", " << line->get_name() << ")" << endl;

  SGEdge *edge=node->first_edge();
  while(edge) {
	SGNode *target=edge->get_target();

	if(edge->get_type()==SGEdge::normal) {
	  os << "       -> " << edge->get_weight() << " to " << (const char *)target->get_station()->get_name() << endl;
	}
	if(edge->get_type()==SGEdge::changeToLine) {
	  Line *targetLine=target->get_line();
	  os << "   ctl -> " << edge->get_realweight() << "/" << edge->get_weight() << " to "
		 << (const char *)targetLine->get_name() << " (" 
		 << (const char *)target->get_station()->get_name() << ")" << endl;
	}
	if(edge->get_type()==SGEdge::changeToStation) {
	  Station *targetStation=target->get_station();
	  os << "   cts -> " << edge->get_realweight() << "/" << edge->get_weight() << " to " 
		 << (const char *)targetStation->get_name() << endl;
	}
	edge=node->next_edge();
  }
  return os;
}
