/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#ifndef __SUBWAYGRAPH_H__
#define __SUBWAYGRAPH_H__


#include "baseclasses.h"
#include "Exceptions.h"
#include "SubwayMap.h"
#include "SMError.h"

static const int cCTLFACTOR=2;
static const int cCTSFACTOR=2;


/*	====================================================
	Classes to represent the subway map as a graph. This
	graph will be used by the Dijkstra route finding
	algorithm.
	==================================================== */

const int INFINITE=0x7fffffff;

class SGEdge;

class SGNode : public Object
{
 public:
  enum SGNodeType {
	normal=0,
	startNode,
	destNode
  };

 public:
  SGNode(Station *station, RouteOfLine *routeOfLine, SGNodeType type=normal) : 
	_station(station), _routeOfLine(routeOfLine), _type(type),
	_distance(INFINITE), _prevNode(NULL) {}

  SGNode(Station *station, SGNodeType type) :
	_station(station), _routeOfLine(NULL), _type(type),
	_distance(INFINITE), _prevNode(NULL) {}

  virtual ~SGNode() {
	_edges.cleanup(1);
  }

  Station *get_station() { return _station; }
  RouteOfLine *get_routeOfLine() { return _routeOfLine; }
  Line *get_line() { 
	return (_routeOfLine==NULL) ? NULL : _routeOfLine->get_line(); 
  }

  bool add_edge(SGEdge *edge);
  SGEdge *first_edge() { return (SGEdge *)_edges.first(_it); }
  SGEdge *next_edge() { return (SGEdge *)_edges.next(_it); }
  SGEdge *find_edgeToTarget(SGNode *node);

  bool remove_tempEdges();

  int get_distance() { return _distance; }
  void set_distance(int distance) { _distance=distance; }

  SGNode *get_prevNode() { return _prevNode; }
  void set_prevNode(SGNode *node) { _prevNode=node; }

  SGNodeType get_type() { return _type; }

 protected:
  // station related with this node
  Station *_station;
  RouteOfLine *_routeOfLine;
  SGNodeType _type;

  int _distance;
  SGNode *_prevNode;

 protected:
  // edges to other nodes
  ObjectListIterator _it;
  ObjectList _edges;


 public:
  friend ostream& operator<<(ostream& os, SGNode *node);

};

class SGEdge : public Object
{
 public:
  enum SGEdgeType {
	normal=0,
	changeToLine,
	changeToStation,
	fromStartNode,
	toDestNode
  };

 public:
  SGEdge(unsigned int weight, SGNode *target, SGEdgeType type=normal) :
	_realweight(weight), _target(target), _type(type) 
	{
	  set_weight(weight);
	}

  virtual ~SGEdge() {}

  unsigned int get_weight() { return _weight; }
  void set_weight(unsigned int value, bool withfactor=true);

  unsigned int get_realweight() { return _realweight; }

  SGNode *get_target() { return _target; }
  SGEdgeType get_type() { return _type; }

 protected:
  unsigned int _weight;
  unsigned int _realweight;
  SGNode *_target;
  SGEdgeType _type;
};


class SubwayGraph : public Object
{
 public:
  enum SGSettings {
	shortest=0,
	direct,
	dontwalk
  };
 public:
  SubwayGraph(SubwayMap *map) : _subwayMap(map), _nodeArray(NULL),
	_ctlFactor(cCTLFACTOR), _ctsFactor(cCTSFACTOR) 
  {
	// Initialize operation day and time to now

	OperationTime::convert_time_t(_opDay, _opHour);
	
  }
  ~SubwayGraph();

  smerr_t build();
  smerr_t cleanup();
  smerr_t reset(SGSettings settings=shortest);

  int add_startNodeForStation(Station *station);
  int add_destNodeForStation(Station *station);
  smerr_t remove_tempNodes();

  int get_numNodes() {
	return (_nodeArray==NULL) ? 0 : _nodeArray->length();
  }

  SGNode *get_nodeAt(int index) { 
	return (_nodeArray==NULL) ? NULL : (SGNode *)_nodeArray->get_at(index); 
  }


  void set_operationTime(OperationTime::EDay day, unsigned int hour) { 
	_opDay=day;
	_opHour=hour; 
  }

 private:
  smerr_t create_nodes();
  smerr_t create_edges();
  smerr_t process_routeOfLine(RouteOfLine *routeOfLine);
  smerr_t add_ctlEdges(SGNode *node, ChangeToLine *ctl);
  smerr_t add_ctsEdges(SGNode *node, ChangeToStation *cts);

  SGNode *find_nodeWithStationOnRoute(Station *station, RouteOfLine *routeOfLine);

 protected:
  SubwayMap *_subwayMap;
  ObjectArray *_nodeArray;

  int _ctlFactor, _ctsFactor;

 private:
  // Operation time
  OperationTime::EDay _opDay;
  unsigned int _opHour;

 public:
  friend ostream& operator<<(ostream& os, SubwayGraph* s);
};


#endif
