/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#ifndef __SUBWAYMAP_H__
#define __SUBWAYMAP_H__

#include "baseclasses.h"
#include "SMError.h"
#include "SMXMLWriter.h"


enum eDirection {
  both=0,			// search in both directions
  forward,			// search only forward or found in forward direction
  backwards,		// search only backwards or found in backwards direction
  not_found			// error in finding something
};

/*	===============================================
	Classes for the internal storage and usage
	of subway maps.
	=============================================== */


/*	===============================================
	Base class for all subway map objects
	=============================================== */

class SubwayMapObject : public Object
{
 public:
  virtual ~SubwayMapObject() {}


  // ----------------------------------------------
  // interface methods
  // should be implemented by each class of the
  // subway map
  // ----------------------------------------------

  // dump the contents as XML
  virtual smerr_t dumpXML() { return 0; }

  virtual void cleanup()=0;

 protected:
  SubwayMapObject() {}
};



class Line;
class Station;

/**
 * Root class of the subway map. Contains stations and lines
 * 
 */

class SubwayMap : public SubwayMapObject
{
 public:
  SubwayMap() {}
  virtual ~SubwayMap() {
	cleanup();
  }

 public:
  void set_name(String& name) { _name=name; }
  void set_city(String& city) { _city=city; }
  void set_descr(String& descr) { _descr=descr; }
  void set_version(String& version) { _version=version; }
  void set_created(String& created) { _created=created; }
  void set_last_change(String& last_change) { _last_change=last_change; }
  void set_created_by(String& created_by) { _created_by=created_by; }

  String& get_name() { return _name; }
  String& get_city() { return _city; }
  String& get_descr() { return _descr; }
  String& get_version() { return _version; }
  String& get_created() { return _created; }
  String& get_last_change() { return _last_change; }
  String& get_created_by() { return _created_by; }

  int get_numStations() { return _stations.length(); }
  int get_numLines() { return _lines.length(); }

  int get_numStationsOnAllLines();

 public:
  // ---------------------------
  // public interface methods
  // ---------------------------

  smerr_t add_station(Station *station);

  Station *first_station(ObjectListIterator& it);
  Station *next_station(ObjectListIterator& it);
  Station *find_stationById(String& id);
  Station *find_stationByName(String& name, ObjectListIterator& it);
  Station *find_stationLikeName(String& name, ObjectListIterator& it);
  smerr_t sort_stations();
  void remove_station(Station *station);


  smerr_t add_line(Line *line);
  Line *first_line(ObjectListIterator& it);
  Line *next_line(ObjectListIterator& it);
  Line *find_lineById(String& id);
  Line *find_lineByName(String& name);

  smerr_t add_servingLinesToStations();


  virtual void cleanup();


  // -----------------------------
  // helper methods
  // -----------------------------
  Line *get_firstLineOfStation(Station *st, ObjectListIterator& it);
  Line *get_nextLineOfStation(Station *st, ObjectListIterator& it);

 public:
  // -----------------------------
  // dump contents
  // -----------------------------
  friend ostream& operator<<(ostream& os, SubwayMap& s);
  friend ostream& operator<<(ostream& os, SubwayMap* s);

  friend SMXMLWriter& operator<<(SMXMLWriter& os, SubwayMap& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, SubwayMap* s);

 protected:
  // ---------------------------
  // subway map information
  // ---------------------------
  String _name;
  String _city;
  String _descr;
  String _version;
  String _created;
  String _last_change;
  String _created_by;

  // ---------------------------
  // reference lists
  // ---------------------------
  ObjectList _stations;
  ObjectList _lines;

};

/*	==============================================
	Station list sorting classes
	============================================== */

class SortStationsByName : public ObjectListSorter
{
 public:
  virtual int compareObjects(Object *route1, Object *route2);
};


/*	================================================
	Classes for the definition of a station in a 
	subway map.
	================================================ */

class ServingLine;

class Station : public SubwayMapObject
{
 public:
  Station() {}
  Station(String& name, String& id) : _name(name), _id(id) {}
  virtual ~Station() {
	cleanup();
  }

  void set_name(String& name) 	{ _name=name; }
  void set_id(String& id) 		{ _id=id; }

  String& get_name() 	{ return _name; }
  String& get_id() 		{ return _id; }

  void add_servingLine(ServingLine *line);

  ServingLine *first_servingLine(ObjectListIterator& it);
  ServingLine *next_servingLine(ObjectListIterator& it);
  int has_servingLine(ServingLine& testLine);
  int has_servingLine(ServingLine *testLine);
  int has_servingLineId(String& id);

  ServingLine *match_servingLine(Station *station);

  int operator==(Station& rval) {
	if(_id==rval._id) return 1;
	return 0;
  }
  int operator==(Station *rval) {
	if(_id==rval->_id) return 1;
	return 0;
  }

  int operator!=(Station& rval) {
	if(_id!=rval._id) return 1;
	return 0;
  }
  int operator!=(Station *rval) {
	if(_id!=rval->_id) return 1;
	return 0;
  }


  virtual void cleanup();

 protected:
  String _name;
  String _id;

 protected:
  ObjectList _servingLines;


 public:
  // -----------------------------
  // dump contents
  // -----------------------------
  friend ostream& operator<<(ostream& os, Station& s);
  friend ostream& operator<<(ostream& os, Station *s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, Station& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, Station* s);

};

class ServingLine : public SubwayMapObject
{
 public:
  ServingLine() {}
  ServingLine(String& id) : _id(id) {}
  virtual ~ServingLine() {
	cleanup();
  }

  void set_id(String& id) { _id=id; }
  String& get_id() { return _id; }

  int operator==(ServingLine& rval) {
	return _id==rval._id;
  }
  
  virtual void cleanup() {
	_id=NULL;
  };

 protected:
  String _id;

 public:
  // -----------------------------
  // dump contents
  // -----------------------------
  friend ostream& operator<<(ostream& os, ServingLine& s);
  friend ostream& operator<<(ostream& os, ServingLine* s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, ServingLine& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, ServingLine* s);

};



/*	================================================
	Classes for the definition of a line in a subway
	map.
	================================================ */

class RouteOfLine;
class StopAt;

class OperationTime : public SubwayMapObject
{
 public:
  enum EDay {
	monday=0,
	tuesday,
	wednesday,
	thursday,
	friday,
	saturday,
	sunday,	
	everyday
  };

 public:
  OperationTime() {
	set_always(everyday);
  }
  ~OperationTime() {}

  void set_always(EDay day);
  void set_never(EDay day);

  smerr_t set_ophour(EDay day, unsigned int hour, bool operates=true);
  smerr_t set_ophours(EDay day, unsigned int *hours, int numhours, bool operates=true);
  
  bool operates_at(EDay day, unsigned int hour);
  bool operates_on(EDay day);
  bool operates_always() { return _always; }

  virtual void cleanup() { set_never(everyday); }


  // convertion functions
  static bool convert_time_t(EDay& day, unsigned int& hour, time_t t=0);


 protected:
  unsigned int hour_to_bit(unsigned int hour);
  void dayname(EDay day, String& str);
  void hourstr(EDay day, String& str);

 protected:
  // Bitmap of operation times for each day

  unsigned int _optimes[7];
  // Description
  // Days:   0=Monday, 1=Tuesday, 2=Wednesday, 3=Thursday, 4=Friday
  //         5=Saturday, 6=Sunday
  // Hours:  Bit 0=0:00, Bit 1=1:00, Bit 2=2:00, ...
  //         Bit 22=22:00, Bit 23=23:00

  bool _always;
  bool _never;

 public:
  friend ostream& operator<<(ostream& os, OperationTime& s);
  friend ostream& operator<<(ostream& os, OperationTime *s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, OperationTime& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, OperationTime* s);

};


class Line : public SubwayMapObject
{
 public:
  Line() {}
  Line(String& name, String& id, String& cl) : _name(name), _id(id), _class(cl) {}
  virtual ~Line() {
	cleanup();
  }

  void set_name(String& name) { _name=name; }
  void set_id(String& id) { _id=id; }
  void set_class(String& cl) { _class=cl; }

  String& get_name() { return _name; }
  String& get_id() { return _id; }
  String& get_class() { return _class; }

  void add_routeOfLine(RouteOfLine *route);

  // Lines should have only one route...
  RouteOfLine *get_routeOfLine();

  // Old methods for lines with multiple routes
  RouteOfLine *first_routeOfLine(ObjectListIterator& it);
  RouteOfLine *next_routeOfLine(ObjectListIterator& it);
  RouteOfLine *get_routeOfLineWithStation(Station *station);

  virtual void cleanup();

  bool operates_at(OperationTime::EDay day, unsigned int hour);

 protected:
  String _name;
  String _id;
  String _class;

  ObjectList _routesOfLine;

 public:
  // -----------------------------
  // dump contents
  // -----------------------------
  friend ostream& operator<<(ostream& os, Line& s);
  friend ostream& operator<<(ostream& os, Line* s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, Line& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, Line* s);


};

class RouteOfLine : public SubwayMapObject
{
 public:
  RouteOfLine() : _oneway(0), _line(NULL) {}
  RouteOfLine(String& name) : _name(name), _oneway(0), _line(NULL) {}
  virtual ~RouteOfLine() {
	cleanup();
  }

  void set_name(String& name) { _name=name; }
  void set_service_start(String& servstart) { _service_start=servstart; }
  void set_service_end(String& servend) { _service_end=servend; }
  void set_startstation(String& station) { _startstation=station; }
  void set_endstation(String& station) { _endstation=station; }
  void set_oneway(int oneway) { _oneway=oneway; }

  String& get_name() { return _name; }
  String& get_service_start() { return _service_start; }
  String& get_service_end() { return _service_end; }
  String& get_startstation() { return _startstation; }
  String& get_endstation() { return _endstation; }
  int get_oneway() { return _oneway; }

  void add_stop(StopAt *stop);

  StopAt *find_stopByStationId(String& id, ObjectListIterator& it);
  StopAt *find_stopByStation(Station *station, ObjectListIterator& it);
  StopAt *first_stop(ObjectListIterator& it);
  StopAt *next_stop(ObjectListIterator& it);
  StopAt *prev_stop(ObjectListIterator& it);

  bool hasStation(Station *station);

  eDirection get_routeDirection(Station *startStation, Station *destStation);

  void cleanup();

  int get_numStops() { return _stops.length(); }

  void set_line(Line *line) { _line=line; }
  Line *get_line() { return _line; }

  OperationTime& get_optime() { return _optime; }

  bool operates_at(OperationTime::EDay day, unsigned int hour);

 protected:
  String _name;
  String _service_start;
  String _service_end;
  String _startstation;
  String _endstation;
  int    _oneway;

  ObjectListIterator _it_stops;
  ObjectList _stops;

  Line *_line;

  OperationTime _optime;

 public:
  // -----------------------------
  // dump contents
  // -----------------------------
  friend ostream& operator<<(ostream& os, RouteOfLine& s);
  friend ostream& operator<<(ostream& os, RouteOfLine* s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, RouteOfLine& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, RouteOfLine* s);

};


class ChangeToLine;
class ChangeToStation;

class StopAt : public SubwayMapObject
{
 public:
  StopAt() : _timeToNext(0), _timeToPrev(0), _station(NULL), _change(0) {}
  StopAt(Station *station, int ttn, int ttp) : _timeToNext(ttn), _timeToPrev(ttp), _station(station),  _change(0) {}
  virtual ~StopAt() {
	cleanup();
  }

  void set_timeToNext(unsigned int ttn) { _timeToNext=ttn; }
  void set_timeToPrev(unsigned int ttp) { _timeToPrev=ttp; }
  int get_timeToNext() { return _timeToNext; }
  int get_timeToPrev() { return _timeToPrev; }

  void set_station(Station *station) { _station=station; }
  Station *get_station() { return _station; }

  void add_changeToLine(ChangeToLine *ctl);
  ChangeToLine *first_changeToLine(ObjectListIterator& it);
  ChangeToLine *next_changeToLine(ObjectListIterator& it);

  void add_changeToStation(ChangeToStation *cts);
  ChangeToStation *first_changeToStation(ObjectListIterator& it);
  ChangeToStation *next_changeToStation(ObjectListIterator& it);

  void set_change(int change) { _change=change; }
  int  get_change() { return _change; }
  void cleanup() {
	_changeToLine.cleanup(1);
	_changeToStation.cleanup(1);
  }
 protected:
  int _timeToNext;		// Time to next station (seconds)
  int _timeToPrev;		// Time to previous station (way back, seconds)
  Station *_station;

  ObjectList _changeToLine;
  ObjectList _changeToStation;
  int _change;

 public:
  // -----------------------------
  // dump contents
  // -----------------------------
  friend ostream& operator<<(ostream& os, StopAt& s);
  friend ostream& operator<<(ostream& os, StopAt* s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, StopAt& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, StopAt* s);

};


class ChangeToLine : public SubwayMapObject
{
 public:
  ChangeToLine() : _id(NULL), _fromDir(both), _toDir(both), _timeToChange(0) {}
  ChangeToLine(String& id, eDirection fromDir=both, eDirection toDir=both, int ttc=0) : 
	_id(id), _fromDir(fromDir), _toDir(toDir), _timeToChange(ttc) {}
  virtual ~ChangeToLine() {}

  void set_id(String& id) { _id=id; }
  void set_fromDir(eDirection dir) { _fromDir=dir; }
  void set_toDir(eDirection dir) { _toDir=dir; }
  void set_timeToChange(int ttc) { _timeToChange=ttc; }

  String& get_id() { return _id; }
  eDirection get_fromDir() { return _fromDir; }
  eDirection get_toDir() { return _toDir; }
  int get_timeToChange() { return _timeToChange; }

  void cleanup() {
	_id=NULL;
  }
 protected:
  String _id;
  eDirection _fromDir;
  eDirection _toDir;
  int _timeToChange;

  friend SMXMLWriter& operator<<(SMXMLWriter& os, ChangeToLine& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, ChangeToLine* s);

};

class ChangeToStation : public SubwayMapObject
{
 public:
  ChangeToStation() : _id(NULL), _fromDir(both), _toDir(both), _timeToChange(0) {}
  ChangeToStation(String& id, eDirection fromDir=both, eDirection toDir=both, int ttc=0) : 
	_id(id), _fromDir(fromDir), _toDir(toDir), _timeToChange(ttc) {}
  virtual ~ChangeToStation() {}

  void set_id(String& id) { _id=id; }
  void set_fromDir(eDirection dir) { _fromDir=dir; }
  void set_toDir(eDirection dir) { _toDir=dir; }
  void set_timeToChange(int ttc) { _timeToChange=ttc; }

  String& get_id() { return _id; }
  eDirection get_fromDir() { return _fromDir; }
  eDirection get_toDir() { return _toDir; }
  int get_timeToChange() { return _timeToChange; }

  void cleanup() {
	_id=NULL;
  }
 protected:
  String _id;
  eDirection _fromDir;
  eDirection _toDir;
  int _timeToChange;

  friend SMXMLWriter& operator<<(SMXMLWriter& os, ChangeToStation& s);
  friend SMXMLWriter& operator<<(SMXMLWriter& os, ChangeToStation* s);

};


#endif
