/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#include "baseclasses.h"

#include <string.h>


/*
  Simple ObjectList
*/


ObjectList::ObjectList() : _numElements(0), _anchor(NULL), _last(NULL)
{

}

ObjectList::~ObjectList()
{

}


Object *ObjectList::add(Object *obj)
{
  ObjectListContainer *tmp=NULL;

  if(_anchor==NULL) {
	// No object in list, store object as anchor
	_anchor=new ObjectListContainer(obj);
	_anchor->set_prev(NULL);
	_anchor->set_next(NULL);
	_last=_anchor;
	_numElements=1;
	return obj;
  }

  // Add object to the end of the list
  tmp=new ObjectListContainer(obj);
  _last->set_next(tmp);
  tmp->set_prev(_last);
  _last=tmp;
  _numElements++;
  
  return _last;
}

Object *ObjectList::addToBegin(Object *obj)
{
  ObjectListContainer *tmp=NULL;

  if(_anchor==NULL) {
	return add(obj);
  }
  
  tmp=new ObjectListContainer(obj);
  tmp->set_next(_anchor);
  tmp->set_prev(NULL);
  _anchor=tmp;
  _numElements++;

  return _anchor;
}

Object *ObjectList::remove(Object *obj)
{
  ObjectListContainer *tmp=NULL;

  tmp=findContainerWithObject(obj);
  if(!tmp) return NULL;

  if(tmp==_anchor) {
	// remove first object
	_anchor=_anchor->get_next();
	if(!_anchor) {
	  // No more objects in list
	  _last=NULL;
	  return obj;
	}

	_anchor->set_prev(NULL);
	delete tmp;
	_numElements--;
	return obj;
  }

  if(tmp==_last) {
	// remove last object
	_last=_last->get_prev();
	if(!_last) {
	  // No more objects in list
	  _anchor=NULL;
	  return obj;
	}
	_last->set_next(NULL);
	delete tmp;
	_numElements--;
	return obj;
  }
  
  // Remove object in the middle
  ObjectListContainer *n=tmp->get_next();
  ObjectListContainer *p=tmp->get_prev();
  n->set_prev(p);
  p->set_next(n);

  delete tmp;
  _numElements--;
  return obj;
}

Object *ObjectList::first(ObjectListIterator& it)
{
  // No first object
  if(!_anchor) return NULL;

  it.set_current(_anchor, this);
  return _anchor->get_object();
}

Object *ObjectList::last(ObjectListIterator& it)
{
  // No last object
  if(!_last) return NULL;

  it.set_current(_last, this);
  return _last->get_object();
}


Object *ObjectList::next(ObjectListIterator& it)
{
  ObjectListContainer *tmp;

  tmp=it.get_current();
  if(!tmp) return NULL;

  tmp=tmp->get_next();
  it.set_current(tmp, this);
  
  if(!tmp) return NULL;

  return tmp->get_object();
}


Object *ObjectList::prev(ObjectListIterator& it)
{
  ObjectListContainer *tmp=NULL;

  tmp=it.get_current();
  if(!tmp) return NULL;

  tmp=tmp->get_prev();
  it.set_current(tmp, this);

  if(!tmp) return NULL;
  return tmp->get_object();
}

void ObjectList::cleanup(int deleteObjects)
{
  ObjectListContainer *tmp=_anchor;

  while(_anchor) {
	if(deleteObjects) {
	  if(_anchor->get_object()) delete _anchor->get_object();
	}
	tmp=_anchor->get_next();
	delete _anchor;

	_anchor=tmp;
  }

  _numElements=0;
}


void ObjectList::bsort(ObjectListSorter& sorter)
{
  ObjectListContainer *cont=NULL;
  int rc;

  for(unsigned int i=0; i<length(); i++) { 	// first loop
	int switched=0;
	cont=_anchor;
	while(cont!=_last) {	// second loop
	  Object *obj1=cont->get_object();
	  Object *obj2=cont->get_next()->get_object();
	  // Compare objects stored within the container
	  rc=sorter.compareObjects(obj1, obj2);

	  // Object 
	  if(rc>0) {
		cont->set_object(obj2);
		cont->get_next()->set_object(obj1);
		switched++;
	  }
	  cont=cont->get_next();
	} // while(second loop)
	if(switched==0) { // no object switched
	  // list is in order
	  return;
	}
  } // for(first loop)
}


ObjectListContainer *ObjectList::findContainerWithObject(Object *obj)
{
  ObjectListContainer *tmp=_anchor;

  while(tmp) {
	if(tmp->get_object()==obj) {
	  // found
	  return tmp;
	}
	tmp=tmp->get_next();
  }

  return NULL;
}

Object *ObjectList::operator[](unsigned int index)
{
  ObjectListContainer *tmp=_anchor;

  if(index>=_numElements) return NULL;

  for(unsigned int i=0; i<index; i++) {
	if(tmp) tmp=tmp->get_next();
  }
  return tmp->get_object();
}


/*  ================================================
	Class ObjectArray
	================================================ */
bool ObjectArray::init(int length)
{
  if(_array) {
	cleanup();
	_array=NULL;
  }

  _array=new ObjectArrayContainer*[length];
  if(_array==NULL) {
	return false;
  }

  for(int i=0; i<length; i++) {
	if((_array[i]=new ObjectArrayContainer)==NULL) {
	  return false;
	}
  }

  _length=length;

  return true;
}

bool ObjectArray::cleanup()
{
  if(_array==NULL) {
	return true;
  }

  for(int i=0; i<_length; i++) {
	if(_array[i]) {
	  delete _array[i];
	  _array[i]=NULL;
	}
  }

  delete [] _array;
  return true;
}

Object *ObjectArray::get_at(int index)
{
  if(index<0 || index>=_length)
	return NULL;

  return (Object *)_array[index]->get_object();
}

bool ObjectArray::set_at(int index, Object *object)
{
  if(index<0 || index>=_length)
	return false;

  _array[index]->set_object(object);
  return true;
}

bool ObjectArray::remove_at(int index)
{
  return set_at(index, NULL);
}

/*  ================================================
	Class String
	================================================ */

String NULLString;


String::String(const String& rval) : _string(NULL)
{

  if(rval._string) {
	_string=new char[strlen(rval._string)+1];
	strcpy(_string, rval._string);
  }
}



String::String(const char *rval) : _string(NULL)
{
  if(rval) {
	_string=new char[strlen(rval)+1];
	strcpy(_string, rval);
  }
}

String& String::operator=(const String& rval)
{
  if(_string) {
	delete [] _string;
	_string=NULL;
  }

  if(rval._string) {
	_string=new char[strlen(rval._string)+1];
	strcpy(_string, rval._string);
  }

  return *this;
}


String& String::operator=(const char *rval)
{
  if(_string) {
	delete [] _string;
	_string=NULL;
  }

  if(rval) {
	_string=new char[strlen(rval)+1];
	strcpy(_string, rval);
  }
  return *this;
}


int String::operator==(const String& rval)
{
  if(!rval._string) return 0;
  if(!_string) return 0;

  return !strcmp(_string, rval._string);
}

int String::operator==(const char *rval)
{
  if(!rval) return 0;
  if(!_string) return 0;

  return !strcmp(_string, rval);
}

int String::operator!=(const String& rval)
{
  if(!rval._string) return 1;
  if(!_string) return 1;

  return strcmp(_string, rval._string);
}

int String::operator!=(const char *rval)
{
  if(!rval) return 1;
  if(!_string) return 1;

  return strcmp(_string, rval);
}

String& String::operator+=(const String& rval)
{
  append(rval._string);
  return *this;
}

String& String::operator+=(const char *rval)
{
  append(rval);
  return *this;
}

String& String::operator+(const String& rval)
{
  append(rval._string);
  return *this;
}

String& String::operator+(const char *rval)
{
  append(rval);
  return *this;
}

bool String::operator<(const String& rval)
{
  if(strcmp(_string, rval._string)<0) return true;
  return false;
}

bool String::operator>(const String& rval)
{
  if(strcmp(_string, rval._string)>0) return true;
  return false;
}

char String::operator[](unsigned int index)
{
  if(index>strlen(_string))
	return 0;	

  return _string[index];
}

int String::append(const char *rval)
{
  char *string=NULL;
  int rvallen=0;
  int newlen=0;
  
  if(!rval) return length();

  rvallen=strlen(rval);

  newlen=length()+rvallen+1;
  string=new char[newlen];
  memset(string, 0, newlen);
  if(_string) {
	strcpy(string, _string);
  }
  strcat(string, rval);

  if(_string) {
	delete [] _string;
	_string=NULL;
  }

  _string=string;
  return newlen;
}


int String::strnicmp(const String& rval, int length)
{
  if(!rval._string || !_string) return -255;

  return ::strncasecmp(_string, rval._string, length);
}
int String::strnicmp(const char *rval, int length)
{
  if(!rval || !_string) return -255;

  return ::strncasecmp(_string, rval, length);
}



ostream& operator<<(ostream& os, const String& s)
{
  return os << s._string;
}
