/*
##########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
##########################################################################
*/



/*
  Base classes for all purposes, e.g. lists, arrays, ...
*/

#ifndef __BASECLASSES_H__
#define __BASECLASSES_H__


// Standard includes

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <stdio.h>

#include <iostream.h>

#ifdef MPATROL
#include <mp/mpatrol.h>
#endif

class ObjectList;
class ObjectListIterator;
class ObjectListContainer;
class ObjectListSorter;

/*	=======================================
	Object - Base class
	======================================= */

class Object 
{
 public:
  virtual ~Object() {};


 protected:
  Object() {}

};


/*	=======================================
	A Simple Object List
	======================================= */

class ObjectList : public Object
{
 public:
  ObjectList();
  ~ObjectList();

  Object *add(Object *obj);
  Object *addToBegin(Object *obj);
  Object *remove(Object *obj);

  Object *first(ObjectListIterator& it);
  Object *last(ObjectListIterator& it);

  Object *next(ObjectListIterator& it);
  Object *prev(ObjectListIterator& it);

  unsigned int length() { return _numElements; }

  Object *operator[] (unsigned int index);

  void cleanup(int deleteObjects=0);

  void bsort(ObjectListSorter& sorter);

 protected:
  ObjectListContainer *findContainerWithObject(Object *obj);
  unsigned int _numElements;

 protected:
  ObjectListContainer *_anchor;
  ObjectListContainer *_last;
};


class ObjectListIterator : public Object
{
 public:
  ObjectListIterator() : Object(), _theList(NULL), _current(NULL) {}

  ObjectListIterator(ObjectListIterator& rval) : Object(), _theList(NULL), _current(NULL) {
	_theList=rval._theList;
	_current=rval._current;
  }

  ~ObjectListIterator() {}

  ObjectListContainer *get_current() { return _current; }
  void set_current(ObjectListContainer *cont, ObjectList *objList) {
	_current=cont;
	_theList=objList;
  }

  ObjectListIterator& operator=(ObjectListIterator& rval) {
	_theList=rval._theList;
	_current=rval._current;
	return *this;
  }
 protected:
  ObjectList *_theList;
  ObjectListContainer *_current;
};


class ObjectListSorter : public Object
{
 public:
  virtual int compareObjects(Object *object1, Object *object2)=0;
};


class ObjectListContainer : public Object
{
 public:
  ObjectListContainer() : _object(NULL), _prev(NULL), _next(NULL) {}
  ObjectListContainer(Object *obj) : _object(obj), _prev(NULL), _next(NULL) {}
  ~ObjectListContainer() {}

  void set_next(ObjectListContainer *cont) { _next=cont; }
  void set_prev(ObjectListContainer *cont) { _prev=cont; }
  ObjectListContainer *get_next() { return _next; }
  ObjectListContainer *get_prev() { return _prev; }

  Object *get_object() { return _object; }
  void    set_object(Object *object) { _object=object; }

 protected:
  Object *_object;
  ObjectListContainer *_prev, *_next;
};


/*	=============================================
	A Simple object array
	============================================= */


class ObjectArrayContainer : public Object
{
 public:
  ObjectArrayContainer() : _object(NULL) {}
  virtual ~ObjectArrayContainer() {}

  Object *get_object() { return _object; }
  void set_object(Object *object) { _object=object; }

 protected:
  Object *_object;
};


class ObjectArray : public Object
{
 public:
  ObjectArray(int length) : _length(0), _array(NULL) {
	init(length);
  }

  virtual ~ObjectArray() {
	cleanup();
  }

  int length() { return _length; }

  Object *get_at(int index);
  bool set_at(int index, Object *object);
  bool remove_at(int index);

 protected:
  int _length;
  ObjectArrayContainer **_array;

  bool init(int length);
  bool cleanup();
};

/*	=============================================
	A Simple string class
	============================================= */

class String : public Object
{
 public:
  String() : _string(NULL) {};
  String(const String& rval);
  String(const char *rval);
  ~String() {
	// cout << this << "::~String() " << _string << endl;
	if(_string) 
	  delete [] _string; 
  }

  unsigned int length() {
	if(!_string) return 0;
	return strlen(_string);
  }

  int to_int() { 
	if(_string) return atoi(_string);
	else return 0;
  }

  int isNULL() {
	if(!_string) return 1;
	return 0;
  }

  String& operator=(const String& rval);
  String& operator=(const char *rval);

  int operator==(const String& rval);
  int operator==(const char *rval);

  int operator!=(const String& rval);
  int operator!=(const char *rval);

  int strnicmp(const String& rval, int length);
  int strnicmp(const char *rval, int length);

  String& operator+=(const String& rval);
  String& operator+=(const char *rval);

  String& operator+(const String& rval);
  String& operator+(const char *rval);

  bool operator<(const String& rval);
  bool operator>(const String& rval);

  char operator[](unsigned int index);

  bool operator!() { 
	if(_string==NULL) return true;
	else return false;
  }
						
  int append(const char *rval);

  operator const char *() { return string(); }
  const char *string() { return (const char *)_string; }

  friend ostream& operator<<(ostream& os, const String& s);

 protected:
  char *_string;
};

extern String NULLString;


#endif
