/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#include "RouteListView.h"
#include "images.h"


// ==== Route list view ===============================================

RouteListView::RouteListView(QWidget *parent, const char *name)
  : QListView(parent, name), _lastItem(NULL), _showPassedStations(true)
{
  setSorting(-1);

  setRootIsDecorated(false);
  setSelectionMode(QListView::NoSelection);
  addColumn("i");
  addColumn("Routes");
}

RouteListView::~RouteListView()
{
}

void RouteListView::cleanup()
{
  clear();
  _lastItem=NULL;
}

void RouteListView::addRoute(Route *route)
{
  if(!route) return;

  QPixmap routeStartPixmap(route_start_xpm);
  QPixmap routeChangePixmap(route_change_xpm);
  QPixmap routeEndPixmap(route_end_xpm);
  QPixmap routeInfoPixmap(route_info_xpm);
  QPixmap routeThroughPixmap(route_through_xpm);
  QPixmap routeByFeet(route_feet_xpm);
  QPixmap routePassedStationPixmap(route_passed_xpm);

  int numChanges=route->get_numConnections()-1;
  int journeyTime=route->calc_journeyTime(1);

  QListViewItem *i=NULL;
  QString tmpStr;

  // --- insert route information to RouteListView ----
  if(numChanges==0) 
	tmpStr.sprintf("direct, %d minutes to go", journeyTime);
  else
	tmpStr.sprintf("%d change%s %d minutes to go", numChanges, ((numChanges==1)?",":"s,"), journeyTime);
  i=new RouteItemInfo(this, routeInfoPixmap, tmpStr);

  // --- Insert connections ----------------------------
  Connection *conn=NULL;
  ObjectListIterator it;
  conn=route->first_connection(it);
  if(conn==NULL) {	// Error, this should not happen
	i=new RouteItemInfo(this, routeInfoPixmap, "no connection");
	return;
  }

  int first=1;
  while(conn) {
	Station *from=conn->get_fromStation();
	Station *to=conn->get_toStation();
	Line *line=conn->get_useLine();
	RouteOfLine *routeOfLine=conn->get_routeOfUsedLine();
	String& dirName=conn->get_directionName();
	int timeToGo=conn->calc_time();
	int timeToChange=conn->get_timeToChange();

	RouteItemStation *s=new RouteItemStation(this, first?routeStartPixmap:routeChangePixmap,
											 (const char *)from->get_name(), 
											 (conn->get_byFeet()?0:timeToChange));
	s->set_fromStation(from);
	s->set_toStation(to);
	s->set_line(line);
	s->set_routeOfUsedLine(routeOfLine);
	if(first) 
		s->set_stationType(RouteItemStation::typeStart);
	first=0;

	if(!conn->get_byFeet()) {
	  // normal U-Bahn connection

	  QPixmap *linePixmap=_iconStore->find_iconByLineId((const char *)line->get_id());
	  RouteItemLine *l=new RouteItemLine(this, routeThroughPixmap, linePixmap,
										 (const char *)line->get_name(),
										 (const char *)dirName,
										 timeToGo);
	  l->set_fromStation(from);
	  l->set_toStation(to);
	  l->set_line(line);
	  l->set_routeOfUsedLine(routeOfLine);
	  
	} else {
	  // connection by feet
	  RouteItemByFeet *l=new RouteItemByFeet(this, routeByFeet, to->get_name(), timeToGo);
	  l->set_fromStation(from);
	  l->set_toStation(to);
	}

	if(_showPassedStations) {
	  Station *passedStation=conn->first_passedStation();
	  while(passedStation) {
		RouteItemPassedStation *p=new RouteItemPassedStation(this, routePassedStationPixmap,
															 (const char *)passedStation->get_name());
		p->set_station(passedStation);															 
		passedStation=conn->next_passedStation();
	  }
	}
	conn=route->next_connection(it);
	if(!conn) {	// end
	  s=new RouteItemStation(this, routeEndPixmap, (const char *)to->get_name(), 0);
	  s->set_fromStation(from);
	  s->set_toStation(to);
	  s->set_line(line);
	  s->set_routeOfUsedLine(routeOfLine);
	  s->set_stationType(RouteItemStation::typeDestination);
	}
  }
  RouteItemSeparator *sep=new RouteItemSeparator(this);
}

// ==== Route list items ===============================================

RouteItemInfo::RouteItemInfo(RouteListView *parent, QPixmap& pixmap, const char *strInfo)
  : RouteItem(parent, parent->getLastItem())
{

  _itemType=RouteItem::typeInfo;
  
  setPixmap(0, pixmap);
  setText(1, QString(strInfo));
  parent->setLastItem(this);
}

RouteItemStation::RouteItemStation(RouteListView *parent, QPixmap& pixmap, const char *stationName, 
								   unsigned int timeToChange)
  : RouteItem(parent, parent->getLastItem()),
    _stationType(typeNormal)
{
	
  _itemType=RouteItem::typeStation;
  
  QString tmpStr=stationName;

  if(timeToChange>0) {  
	  tmpStr+=", ";
	  tmpStr+=QString::number(timeToChange);
	  tmpStr+=" min to change";
  }
  
  setPixmap(0, pixmap);
  setText(1, tmpStr);
  parent->setLastItem(this);
}

RouteItemLine::RouteItemLine(RouteListView *parent, QPixmap& pix0, QPixmap *linePixmap, const char *line, const char *direction,
							 int timeToGo)
  : RouteItem(parent, parent->getLastItem())
  
{
  _itemType=RouteItem::typeLine;


  QString tmpStr;
  if(!linePixmap) // no pixmap
	tmpStr+=line;
  tmpStr+=" [";
  tmpStr+=direction;
  tmpStr+="] ";
  tmpStr+=QString::number(timeToGo);
  tmpStr+=" min";

  setPixmap(0, pix0);
  if(linePixmap) setPixmap(1, *linePixmap);
  setText(1, tmpStr); 

  parent->setLastItem(this);
}

RouteItemByFeet::RouteItemByFeet(RouteListView *parent, QPixmap& pix0, const char *direction, int timeToGo)
  : RouteItem(parent, parent->getLastItem())
    
{
  _itemType=RouteItem::typeByFeet;

  QString tmpStr;
  
  tmpStr="to ";

  if(direction)
	tmpStr+=direction;
  else
	tmpStr+="???";

  tmpStr+=", ";
  tmpStr+=QString::number(timeToGo);
  tmpStr+=" min to walk";

  setPixmap(0, pix0);
  setText(1, tmpStr);

  parent->setLastItem(this);
}

RouteItemSeparator::RouteItemSeparator(RouteListView *parent)
  : RouteItem(parent, parent->getLastItem())
{
  _itemType=RouteItem::typeSeparator;

  //setText(0, " ");
  parent->setLastItem(this);
}

RouteItemPassedStation::RouteItemPassedStation(RouteListView *parent, QPixmap& pixmap, const char *stationName)
  : RouteItem(parent, parent->getLastItem())
    
{
  _itemType=RouteItem::typePassedStation;

  QString tmpString;

  tmpString="  - ";
  tmpString+=QString(stationName);

  setPixmap(0, pixmap);
  setText(1, tmpString);
  parent->setLastItem(this);
}
