/*
#########################################################################
# Subway Map
# ----------
# Copyright (c) 2004 by Dimitri Brukakis
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
#########################################################################
*/

#include "XMLIconFileParser.h"



/*	================================================
	Icon file parser base class
	================================================ */

smerr_t XMLIconFileParser::parse(const char *filename, UBahnIconStore *iconStore)
{
  int rc;

  _state=none;
  _iconStore=iconStore;

  // initialize parser

  rc=initParser();
  if(rc) return rc;

  rc=parseDocument(filename);
  if(rc) return rc;

  return 0;
}

smerr_t XMLIconFileParser::parse_element(XMLElement& e)
{
  
  // ------------------
  // start document
  // ------------------
  if(e.get_name()==tag_icons) {
	if(_state!=none) 
	  return exception(new XMLException((const String&)"invalid tag, already in <icons>",
										smeXML_INVALID_TAG, e.get_name(), _line, _column));
	_state=start;
	return smeOK;
  }

  if(e.get_name()==tag_lineicons) {
	if(_state!=start)
	  return exception(new XMLException((const String&)"invalid tag",
										smeXML_INVALID_TAG, e.get_name(), _line, _column));
	_state=in_lines;
	return smeOK;
  }
   
  if(e.get_name()==tag_icon && _state==in_lines) {
	return parse_lineicon(e);
  }

  return exception(new XMLException((const String&)"unknown element", smeXML_UNKNOWN_TAG,
									e.get_name(), _line, _column));
}

smerr_t XMLIconFileParser::finish_element(XMLElement& e)
{
  if(e.get_name()==tag_icons) {
	_state=finished;
  }

  if(e.get_name()==tag_lineicons) {
	_state=start;
  }

  return smeOK;
}

smerr_t XMLIconFileParser::parse_lineicon(XMLElement& e)
{
  XMLAttribute& aline=e.get_attribute(attr_icon_line);
  XMLAttribute& afile=e.get_attribute(attr_icon_file);
  int rc;

  if(aline==unknownAttribute)
	return exception(new XMLException((const String&)"<icon> missing attribute: line",
									  smeXML_MISSING_ATTR, e.get_name(), _line, _column));

  if(afile==unknownAttribute)
	return exception(new XMLException((const String&)"<icon> missing attribute: file",
									  smeXML_MISSING_ATTR, e.get_name(), _line, _column));

  rc=_iconStore->add_lineIcon((const char *)aline.get_value(), (const char *)afile.get_value());
  if(rc!=smeOK)
	return exception(new FileException((const String&)"File not found",
									   smeFILE_NOT_FOUND, (const char *)afile.get_value()));
  return smeOK;
}



/*	================================================
	Icon file parser Qt implementation
	================================================ */

int XMLIconFileParser_Qt::initParser()
{

  // Nothing to do, yet.
  return smeOK;
}


int XMLIconFileParser_Qt::parseDocument(const char *filename)
{
  bool rc;
  QFile xmlFile(filename);

  if(!xmlFile.exists()) {
	return exception(new FileException((const String)"File not found", smeFILE_NOT_FOUND, filename));
  }

  QXmlInputSource source(xmlFile);
  QXmlSimpleReader reader;

  _locator=new QXmlLocator(&reader);

  //  reader.setLocator(locator);
  reader.setContentHandler(this);

  if((rc=reader.parse(source))==FALSE) {
	// something went wrong

	delete _locator;
	_locator=NULL;

	if(_exception) {
	  return _exception->get_extype();
	}
	// returned FALSE but no exception, maybe parsing error
	return exception(new XMLException((const String&)"parsing error", smeXML_PARSING_ERROR, _line, _column));
  }

  delete _locator;
  _locator=NULL;

  return smeOK;
}



/*	===================================================
	Implementation of Qt SAX interface methods
	=================================================== */

bool XMLIconFileParser_Qt::startDocument()
{
  // cdebug << "startDocument()" << endl;
  return TRUE;
}


bool XMLIconFileParser_Qt::endDocument()
{
  // cdebug << "endDocument()" << endl;

  return TRUE;
}

bool XMLIconFileParser_Qt::startElement(const QString& namespaceURI,
										 const QString& qName, const QString& localName,
										 const QXmlAttributes& attrs)
{
  // cdebug << "startElement()" << endl;

  XMLElement element;
  element.set_name((const char *)localName);

  for(int a=0; a<attrs.length(); a++) {
	XMLString aname((const char *)attrs.localName(a));
	XMLString avalue((const char *)attrs.value(a));

	element.add_attribute(aname, avalue);
  }

  // Store document position information
  set_linecolumn(_locator->lineNumber(), _locator->columnNumber());

  if(parse_element(element)!=smeOK)
	return FALSE;	// error information in exception object

  return TRUE;
}

bool XMLIconFileParser_Qt::endElement(const QString& namespaceURI,
									   const QString& qName, const QString& localName)
{
  // cdebug << "endElement()" << endl;
  XMLElement element;
  element.set_name((const char *)localName);

  // Store document position information
  set_linecolumn(_locator->lineNumber(), _locator->columnNumber());

  if(finish_element(element)!=smeOK)
	return FALSE;
  return TRUE;
}

bool XMLIconFileParser_Qt::characters(const QString& chr)
{

  return TRUE;
}
